"""
Implementation fo Zemel's Learned Fair Representations lifted from AIF360
https://github.com/IBM/AIF360/blob/master/aif360/algorithms/preprocessing/lfr.py
"""
from typing import List, Tuple, Dict, Optional, Union

import pandas as pd
import numpy as np
import scipy.optimize as optim
from numba import jit

from ethicml.utility.data_structures import DataTuple, TestTuple
from ethicml.implementations.utils import (
    load_data_from_flags,
    save_transformations,
    pre_algo_argparser,
)

# Disable pylint's naming convention complaints - this code wasn't implemented by us
# pylint: disable=invalid-name


@jit
def distances(X, v, alpha, N, P, k):
    """
    Calculates distances?
    Args:
        X:
        v:
        alpha:
        N:
        P:
        k:

    Returns:

    """
    dists = np.zeros((N, P))
    for i in range(N):
        for p in range(P):
            for j in range(k):
                dists[i, j] += (X[i, p] - v[j, p]) * (X[i, p] - v[j, p]) * alpha[p]
    return dists


@jit
def M_nk(dists, N, k):
    """
    Not a clue
    Args:
        dists:
        N:
        k:

    Returns:

    """
    matrix_nk = np.zeros((N, k))
    exp = np.zeros((N, k))
    denom = np.zeros(N)
    for i in range(N):
        for j in range(k):
            exp[i, j] = np.exp(-1 * dists[i, j])
            denom[i] += exp[i, j]
        for j in range(k):
            if denom[i]:
                matrix_nk[i, j] = exp[i, j] / denom[i]
            else:
                matrix_nk[i, j] = exp[i, j] / 1e-6
    return matrix_nk


@jit
def M_k(matrix_nk, N, k):
    """
    ???
    Args:
        matrix_nk:
        N:
        k:

    Returns:

    """
    matrix_k = np.zeros(k)
    for j in range(k):
        for i in range(N):
            matrix_k[j] += matrix_nk[i, j]
        matrix_k[j] /= N
    return matrix_k


@jit
def x_n_hat(X, matrix_nk, v, N, P, k):
    """???"""
    x_n_hat_obj = np.zeros((N, P))
    L_x = 0.0
    for i in range(N):
        for p in range(P):
            for j in range(k):
                x_n_hat_obj[i, p] += matrix_nk[i, j] * v[j, p]
            L_x += (X[i, p] - x_n_hat_obj[i, p]) * (X[i, p] - x_n_hat_obj[i, p])
    return x_n_hat_obj, L_x


@jit
def yhat(matrix_nk, y, w, N, k):
    """no idea"""
    y_hat = np.zeros(N)
    L_y = 0.0
    for i in range(N):
        for j in range(k):
            y_hat[i] += matrix_nk[i, j] * w[j]
        y_hat[i] = 1e-6 if y_hat[i] <= 0 else y_hat[i]
        y_hat[i] = 0.999 if y_hat[i] >= 1 else y_hat[i]
        L_y += -1 * y[i] * np.log(y_hat[i]) - (1.0 - y[i]) * np.log(1.0 - y_hat[i])
    return y_hat, L_y


@jit
def yhat_without_loss(matrix_nk, w, N, k):
    """no idea without loss"""
    y_hat = np.zeros(N)
    for i in range(N):
        for j in range(k):
            y_hat[i] += matrix_nk[i, j] * w[j]
        y_hat[i] = 1e-6 if y_hat[i] <= 0 else y_hat[i]
        y_hat[i] = 0.999 if y_hat[i] >= 1 else y_hat[i]
    return y_hat


@jit
def LFR_optim_obj(
    params,
    data_sensitive,
    data_nonsensitive,
    y_sensitive,
    y_nonsensitive,
    k=10,
    A_x=0.01,
    A_y=0.1,
    A_z=0.5,
    results=0,
):
    """
    The funtion to be optimized
    Args:
        params:
        data_sensitive:
        data_nonsensitive:
        y_sensitive:
        y_nonsensitive:
        k:
        A_x:
        A_y:
        A_z:
        results:

    Returns:

    """
    LFR_optim_obj.iters += 1
    Ns, P = data_sensitive.shape
    Nns, _ = data_nonsensitive.shape

    alpha0 = params[:P]
    alpha1 = params[P : 2 * P]
    w = params[2 * P : (2 * P) + k]
    v = np.array(params[(2 * P) + k :]).reshape((k, P))

    dists_sensitive = distances(data_sensitive, v, alpha1, Ns, P, k)
    dists_nonsensitive = distances(data_nonsensitive, v, alpha0, Nns, P, k)

    M_nk_sensitive = M_nk(dists_sensitive, Ns, k)
    M_nk_nonsensitive = M_nk(dists_nonsensitive, Nns, k)

    M_k_sensitive = M_k(M_nk_sensitive, Ns, k)
    M_k_nonsensitive = M_k(M_nk_nonsensitive, Nns, k)

    L_z = 0.0
    for j in range(k):
        L_z += abs(M_k_sensitive[j] - M_k_nonsensitive[j])

    _, L_x1 = x_n_hat(data_sensitive, M_nk_sensitive, v, Ns, P, k)
    _, L_x2 = x_n_hat(data_nonsensitive, M_nk_nonsensitive, v, Nns, P, k)
    L_x = L_x1 + L_x2

    yhat_sensitive, L_y1 = yhat(M_nk_sensitive, y_sensitive, w, Ns, k)
    yhat_nonsensitive, L_y2 = yhat(M_nk_nonsensitive, y_nonsensitive, w, Nns, k)
    L_y = L_y1 + L_y2

    criterion = A_x * L_x + A_y * L_y + A_z * L_z

    return_tuple = (yhat_sensitive, yhat_nonsensitive, M_nk_sensitive, M_nk_nonsensitive)
    return return_tuple if results else criterion


LFR_optim_obj.iters = 0


def train_and_transform(
    train: DataTuple, test: TestTuple, flags: Dict[str, Union[int, float]]
) -> (Tuple[DataTuple, TestTuple]):
    """Train the Zemel model and return the transformed features of the train and test sets"""
    np.random.seed(888)
    features_dim = train.x.shape[1]

    sens_col = train.s.columns[0]
    training_sensitive = train.x.loc[train.s[sens_col] == 0].to_numpy()
    training_nonsensitive = train.x.loc[train.s[sens_col] == 1].to_numpy()
    ytrain_sensitive = train.y.loc[train.s[sens_col] == 0].to_numpy()
    ytrain_nonsensitive = train.y.loc[train.s[sens_col] == 1].to_numpy()

    model_inits = np.random.uniform(
        size=int(features_dim * 2 + flags['clusters'] + features_dim * flags['clusters'])
    )
    bnd: List[Tuple[Optional[int], Optional[int]]] = []
    for i, _ in enumerate(model_inits):
        if i < features_dim * 2 or i >= features_dim * 2 + flags['clusters']:
            bnd.append((None, None))
        else:
            bnd.append((0, 1))

    learned_model = optim.fmin_l_bfgs_b(
        LFR_optim_obj,
        x0=model_inits,
        epsilon=flags['epsilon'],
        args=(
            training_sensitive,
            training_nonsensitive,
            ytrain_sensitive[:, 0],
            ytrain_nonsensitive[:, 0],
            flags['clusters'],
            flags['Ax'],
            flags['Ay'],
            flags['Az'],
            0,
        ),
        bounds=bnd,
        approx_grad=True,
        maxfun=flags['maxfun'],
        maxiter=flags['max_iter'],
        disp=False,
    )[0]

    testing_sensitive = test.x.loc[test.s[sens_col] == 0].to_numpy()
    testing_nonsensitive = test.x.loc[test.s[sens_col] == 1].to_numpy()
    # ytest_sensitive = test.y.loc[test.s[sens_col] == 0].to_numpy()
    # ytest_nonsensitive = test.y.loc[test.s[sens_col] == 1].to_numpy()

    # Mutated, fairer dataset with new labels
    test_transformed = transform(
        testing_sensitive, testing_nonsensitive, learned_model, test, flags
    )

    training_sensitive = train.x.loc[train.s[sens_col] == 0].to_numpy()
    training_nonsensitive = train.x.loc[train.s[sens_col] == 1].to_numpy()
    ytrain_sensitive = train.y.loc[train.s[sens_col] == 0].to_numpy()
    ytrain_nonsensitive = train.y.loc[train.s[sens_col] == 1].to_numpy()

    # extract training model parameters
    train_transformed = transform(
        training_sensitive, training_nonsensitive, learned_model, train, flags
    )

    return (
        DataTuple(x=train_transformed, s=train.s, y=train.y, name=train.name),
        TestTuple(x=test_transformed, s=test.s, name=test.name),
    )


def transform(features_sens, features_nonsens, learned_model, dataset, flags):
    """
    transform a dataset based on the
    Args:
        features_sens:
        features_nonsens:
        label_sens:
        label_nonsens:
        learned_model:
        dataset:

    Returns:

    """
    k = flags['clusters']
    Ns, P = features_sens.shape
    N, _ = features_nonsens.shape
    alphaoptim0 = learned_model[:P]
    alphaoptim1 = learned_model[P : 2 * P]
    # woptim = learned_model[2 * P : (2 * P) + k]
    voptim = np.array(learned_model[(2 * P) + k :]).reshape((k, P))

    # compute distances on the test dataset using train model params
    dist_sensitive = distances(features_sens, voptim, alphaoptim1, Ns, P, k)
    dist_nonsensitive = distances(features_nonsens, voptim, alphaoptim0, N, P, k)

    # compute cluster probabilities for test instances
    M_nk_sensitive = M_nk(dist_sensitive, Ns, k)
    M_nk_nonsensitive = M_nk(dist_nonsensitive, N, k)

    # learned mappings for test instances
    res_sensitive = x_n_hat(features_sens, M_nk_sensitive, voptim, Ns, P, k)
    x_n_hat_sensitive = res_sensitive[0]
    res_nonsensitive = x_n_hat(features_nonsens, M_nk_nonsensitive, voptim, N, P, k)
    x_n_hat_nonsensitive = res_nonsensitive[0]

    # compute predictions for test instances
    # y_hat_sensitive = yhat_without_loss(M_nk_sensitive, woptim, Ns, k)
    # y_hat_nonsensitive = yhat_without_loss(M_nk_nonsensitive, woptim, N, k)

    sens_col = dataset.s.columns[0]

    sensitive_idx = dataset.x[dataset.s[sens_col] == 0].index
    nonsensitive_idx = dataset.x[dataset.s[sens_col] == 1].index

    transformed_features = np.zeros_like(dataset.x.values)
    # transformed_labels = np.zeros_like(dataset.y.values)
    transformed_features[sensitive_idx] = x_n_hat_sensitive
    transformed_features[nonsensitive_idx] = x_n_hat_nonsensitive
    # transformed_labels[sensitive_idx] = np.reshape(y_hat_sensitive, [-1, 1])
    # transformed_labels[nonsensitive_idx] = np.reshape(y_hat_nonsensitive, [-1, 1])
    # transformed_labels = (np.array(transformed_labels) > flags['threshold']).astype(np.float64)

    # return DataTuple(
    #     x=pd.DataFrame(transformed_features, columns=dataset.x.columns),
    #     s=pd.DataFrame(dataset.s, columns=dataset.s.columns),
    #     y=pd.DataFrame(transformed_labels, columns=dataset.y.columns),
    # )

    return pd.DataFrame(transformed_features, columns=dataset.x.columns)


def main():
    """main method to run model"""
    parser = pre_algo_argparser()
    parser.add_argument("--clusters", type=int, required=True)
    parser.add_argument("--Ax", type=float, required=True)
    parser.add_argument("--Ay", type=float, required=True)
    parser.add_argument("--Az", type=float, required=True)
    parser.add_argument("--max_iter", type=int, required=True)
    parser.add_argument("--maxfun", type=int, required=True)
    parser.add_argument("--epsilon", type=float, required=True)
    parser.add_argument("--threshold", type=float, required=True)
    args = parser.parse_args()
    flags = vars(parser.parse_args())

    train, test = load_data_from_flags(flags)
    save_transformations(train_and_transform(train, test, flags), args)


if __name__ == "__main__":
    main()



"""
Variational Fair Auto-Encoder by Louizos et al
"""

from pathlib import Path
from typing import List, Tuple, Dict, Union, Optional

from ethicml.algorithms.preprocess.pre_algorithm import PreAlgorithmAsync
from ethicml.algorithms.preprocess.interface import flag_interface
from ethicml.utility.data_structures import PathTuple, TestPathTuple, DataTuple, TestTuple


class VFAE(PreAlgorithmAsync):
    """
    VFAE Object - see implementation file for details
    """

    def __init__(
        self,
        dataset: str,
        supervised: bool = True,
        epochs: int = 10,
        batch_size: int = 32,
        fairness: str = "DI",
        z1_enc_size: Optional[List[int]] = None,
        z2_enc_size: Optional[List[int]] = None,
        z1_dec_size: Optional[List[int]] = None,
    ):
        # pylint: disable=too-many-arguments
        super().__init__()

        if z1_enc_size is None:
            z1_enc_size = [100]
        if z2_enc_size is None:
            z2_enc_size = [100]
        if z1_dec_size is None:
            z1_dec_size = [100]

        self.flags: Dict[str, Union[int, str, List[int]]] = {
            "supervised": supervised,
            "fairness": fairness,
            "batch_size": batch_size,
            "epochs": epochs,
            "dataset": dataset,
            "z1_enc_size": z1_enc_size,
            "z2_enc_size": z2_enc_size,
            "z1_dec_size": z1_dec_size,
        }

    def run(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        from ...implementations.vfae import train_and_transform

        return train_and_transform(train, test, self.flags)

    def _script_command(
        self,
        train_paths: PathTuple,
        test_paths: TestPathTuple,
        new_train_x_path: Path,
        new_train_s_path: Path,
        new_train_y_path: Path,
        new_train_name_path: Path,
        new_test_x_path: Path,
        new_test_s_path: Path,
        new_test_name_path: Path,
    ) -> List[str]:
        args = flag_interface(
            train_paths,
            test_paths,
            new_train_x_path,
            new_train_s_path,
            new_train_y_path,
            new_train_name_path,
            new_test_x_path,
            new_test_s_path,
            new_test_name_path,
            self.flags,
        )
        return ["-m", "ethicml.implementations.vfae"] + args

    @property
    def name(self) -> str:
        return "VFAE"



"""
Wrapper for SKLearn implementation of SVM
"""
from typing import Optional

from sklearn.svm import SVC

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmAsync
from ethicml.algorithms.inprocess.interface import conventional_interface
from ethicml.implementations import svm


class SVM(InAlgorithmAsync):
    """Support Vector Machine"""

    def __init__(self, C: Optional[float] = None, kernel: Optional[str] = None):
        super().__init__()
        self.C = SVC().C if C is None else C
        self.kernel = SVC().kernel if kernel is None else kernel

    def run(self, train, test):
        return svm.train_and_predict(train, test, self.C, self.kernel)

    def _script_command(self, train_paths, test_paths, pred_path):
        script = ["-m", svm.train_and_predict.__module__]
        args = conventional_interface(
            train_paths, test_paths, pred_path, str(self.C), str(self.kernel)
        )
        return script + args

    @property
    def name(self) -> str:
        return "SVM"



"""
This module contains items related to data, such as raw csv's and data objects
"""

from . import load, configurable_dataset, dataset
from .adult import Adult
from .compas import Compas
from .credit import Credit
from .german import German
from .sqf import Sqf
from .toy import Toy
from .non_binary_toy import NonBinaryToy



"""
Wrapper for calling the fair GP model
"""
from pathlib import Path
from tempfile import TemporaryDirectory

from typing import List
import numpy as np
import pandas as pd

from ethicml.algorithms.inprocess.installed_model import InstalledModel
from ethicml.utility.data_structures import DataTuple, TestTuple

PRED_FNAME = "predictions.npz"
MAX_EPOCHS = 1000
MAX_BATCH_SIZE = 10100  # can go up to 10000
MAX_NUM_INDUCING = 5000  # 2500 seems to be more than enough
SEED = 888


class GPyT(InstalledModel):
    """
    Normal GP model
    """

    basename = "GPyT"

    def __init__(self, s_as_input=True, gpu=0, epochs=70, length_scale=1.2):
        super().__init__(
            name="gpyt",
            url="https://github.com/predictive-analytics-lab/fair-gpytorch.git",
            module="fair-gpytorch",
            file_name="run.py",
        )
        self.s_as_input = s_as_input
        self.gpu = gpu
        self.epochs = epochs
        self.length_scale = length_scale

    async def run_async(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        (ytrain,), label_converter = _fix_labels([train.y.to_numpy()])
        raw_data = dict(
            xtrain=train.x.to_numpy(),
            strain=train.s.to_numpy(),
            ytrain=ytrain,
            xtest=test.x.to_numpy(),
            stest=test.s.to_numpy(),
            # ytest=ytest,
        )
        parameters = self._additional_parameters(raw_data)

        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            data_path = tmp_path / "data.npz"
            np.savez(data_path, **raw_data)
            model_name = "local"
            flags = _flags(
                parameters,
                str(data_path),
                tmpdir,
                self.s_as_input,
                model_name,
                raw_data["ytrain"].shape[0],
                self.gpu,
                self.epochs,
                self.length_scale,
            )
            await self._run_gpyt(flags)

            # Read the results from the numpy file 'predictions.npz'
            with (tmp_path / model_name / PRED_FNAME).open("rb") as file_obj:
                output = np.load(file_obj)
                pred_mean = output["pred_mean"]

        predictions = label_converter((pred_mean > 0.5).astype(raw_data["ytrain"].dtype)[:, 0])
        return pd.DataFrame(predictions, columns=["preds"])

    async def _run_gpyt(self, flags):
        """Generate command to run GPyT"""
        cmd = [str(self._module_path() / self.file_name)]
        for key, value in flags.items():
            cmd += [f"--{key}", str(value)]
        await self._call_script(cmd)

    @staticmethod
    def _additional_parameters(_):
        return dict(lik="BaselineLikelihood")

    @property
    def name(self):
        return f"{self.basename}_in_{self.s_as_input}"


class GPyTDemPar(GPyT):
    """GP algorithm which enforces demographic parity"""

    MEAN = 2
    MIN = 3
    MAX = 4

    def __init__(
        self,
        target_acceptance=None,
        average_prediction=False,
        target_mode=MEAN,
        marginal=False,
        precision_target=1.0,
        **kwargs,
    ):
        """
        Args:
            s_as_input: should the sensitive attribute be part of the input?
            target_acceptance: which acceptance rate to target
            average_prediction: whether to use to average of all possible sensitive attributes for
                                predictions
            target_mode: if no target rate is given, how is the target chosen?
            marginal: when doing average_prediction, should the prior of s be taken into account?
            precision_target: how similar should target labels and true labels be
        """
        super().__init__(**kwargs)
        if self.s_as_input and average_prediction:
            self.__name = f"{self.basename}_dem_par_av_True"
            if marginal:
                self.__name += "_marg"
        else:
            self.__name = f"{self.basename}_dem_par_in_{self.s_as_input}"
        if target_acceptance is not None:
            self.__name += f"_tar_{target_acceptance}"
        elif target_mode != self.MEAN:
            if target_mode == self.MIN:
                self.__name += "_tar_min"
            elif target_mode == self.MAX:
                self.__name += "_tar_max"
            else:
                self.__name += f"_tar_{target_mode}"
        if precision_target != 1.0:
            self.__name += f"_pt_{precision_target}"
        self.target_acceptance = target_acceptance
        self.target_mode = target_mode
        self.average_prediction = average_prediction
        self.marginal = marginal
        self.precision_target = precision_target

    def _additional_parameters(self, raw_data):
        biased_acceptance = compute_bias(raw_data["ytrain"], raw_data["strain"])

        if self.target_acceptance is None:
            if self.target_mode == self.MEAN:
                target_rate = 0.5 * (biased_acceptance[0] + biased_acceptance[1])
            elif self.target_mode == self.MIN:
                target_rate = min(biased_acceptance[0], biased_acceptance[1])
            elif self.target_mode == self.MAX:
                target_rate = max(biased_acceptance[0], biased_acceptance[1])
            else:
                acc_min = min(biased_acceptance[0], biased_acceptance[1])
                acc_max = max(biased_acceptance[0], biased_acceptance[1])
                target_rate = acc_min + self.target_mode * (acc_max - acc_min)
        else:
            target_rate = self.target_acceptance

        if self.marginal:
            p_s = prior_s(raw_data["strain"])
        else:
            p_s = [0.5] * 2

        return dict(
            lik="TunePrLikelihood",
            target_rate1=target_rate[0] if isinstance(target_rate, tuple) else target_rate,
            target_rate2=target_rate[1] if isinstance(target_rate, tuple) else target_rate,
            biased_acceptance1=biased_acceptance[0],
            biased_acceptance2=biased_acceptance[1],
            probs_from_flipped=False,
            average_prediction=self.average_prediction,
            p_s0=p_s[0],
            p_s1=p_s[1],
            p_ybary0_or_ybary1_s0=self.precision_target,
            p_ybary0_or_ybary1_s1=self.precision_target,
        )

    @property
    def name(self):
        return self.__name


class GPyTEqOdds(GPyT):
    """GP algorithm which enforces equality of opportunity"""

    def __init__(
        self,
        average_prediction=False,
        tpr=None,
        marginal=False,
        tnr0=None,
        tnr1=None,
        tpr0=None,
        tpr1=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if self.s_as_input and average_prediction:
            self.__name = "{self.basename}_eq_odds_av_True"
            if marginal:
                self.__name += "_marg"
        else:
            self.__name = f"{self.basename}_eq_odds_in_{self.s_as_input}"

        self.odds = None
        if any(x is not None for x in [tnr0, tnr1, tpr0, tpr1]):  # if any of them is not `None`
            self.odds = {}
            for val, name, target in [
                (tnr0, "0tnr", "p_ybary0_s0"),
                (tnr1, "1tnr", "p_ybary0_s1"),
                (tpr0, "0tpr", "p_ybary1_s0"),
                (tpr1, "1tpr", "p_ybary1_s1"),
            ]:
                if val is not None:
                    self.odds[target] = val
                    self.__name += f"_{name}_{val}"  # add to name
                else:
                    self.odds[target] = 1.0  # default value
        elif tpr is not None:
            self.odds = dict(p_ybary0_s0=1.0, p_ybary0_s1=1.0, p_ybary1_s0=tpr, p_ybary1_s1=tpr)
            self.__name += f"_tpr_{tpr}"

        self.average_prediction = average_prediction
        self.marginal = marginal

    def _additional_parameters(self, raw_data):
        biased_acceptance = compute_bias(raw_data["ytrain"], raw_data["strain"])

        if self.marginal:
            p_s = prior_s(raw_data["strain"])
        else:
            p_s = [0.5] * 2

        return dict(
            lik="TuneTprLikelihood",
            p_ybary0_s0=1.0,
            p_ybary0_s1=1.0,
            p_ybary1_s0=1.0,
            p_ybary1_s1=1.0,
            biased_acceptance1=biased_acceptance[0],
            biased_acceptance2=biased_acceptance[1],
            average_prediction=self.average_prediction,
            p_s0=p_s[0],
            p_s1=p_s[1],
        )

    async def run_async(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        (ytrain,), label_converter = _fix_labels([train.y.to_numpy()])
        raw_data = dict(
            xtrain=train.x.to_numpy(),
            strain=train.s.to_numpy(),
            ytrain=ytrain,
            xtest=test.x.to_numpy(),
            stest=test.s.to_numpy(),
            # ytest=ytest,
        )
        parameters = self._additional_parameters(raw_data)

        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            data_path = tmp_path / Path("data.npz")
            model_name = "local"  # f"run{self.counter}_s_as_input_{self.s_as_input}"
            flags = _flags(
                parameters,
                str(data_path),
                tmpdir,
                self.s_as_input,
                model_name,
                len(raw_data["ytrain"]),
                self.gpu,
                self.epochs,
                self.length_scale,
            )

            if self.odds is None:
                # Split the training data into train and dev and save it to `data.npz`
                train_dev_data = split_train_dev(
                    raw_data["xtrain"], raw_data["ytrain"], raw_data["strain"]
                )
                np.savez(data_path, **train_dev_data)

                # First run
                await self._run_gpyt(flags)

                # Read the results from the numpy file 'predictions.npz'
                with (tmp_path / model_name / PRED_FNAME).open("rb") as file_obj:
                    output = np.load(file_obj)
                    prediction_on_train = output["pred_mean"]
                preds = (prediction_on_train > 0.5).astype(int)
                odds = compute_odds(train_dev_data["ytest"], preds, train_dev_data["stest"])

                # Enforce equality of opportunity
                opportunity = min(odds["p_ybary1_s0"], odds["p_ybary1_s1"])
                odds["p_ybary1_s0"] = opportunity
                odds["p_ybary1_s1"] = opportunity
                flags.update({"epochs": 2 * flags["epochs"], **odds})
            else:
                flags.update(self.odds)

            # Save with real test data
            np.savez(data_path, **raw_data)

            # Second run
            await self._run_gpyt(flags)

            # Read the results from the numpy file 'predictions.npz'
            with (tmp_path / model_name / PRED_FNAME).open("rb") as file_obj:
                output = np.load(file_obj)
                pred_mean = output["pred_mean"]

        # Convert the result to the expected format
        predictions = label_converter((pred_mean > 0.5).astype(raw_data["ytrain"].dtype)[:, 0])
        return pd.DataFrame(predictions, columns=["preds"])

    @property
    def name(self):
        return self.__name


def prior_s(sensitive):
    """Compute the bias in the labels with respect to the sensitive attributes"""
    return (np.sum(sensitive == 0) / len(sensitive), np.sum(sensitive == 1) / len(sensitive))


def compute_bias(labels, sensitive):
    """Compute the bias in the labels with respect to the sensitive attributes"""
    rate_y1_s0 = np.sum(labels[sensitive == 0] == 1) / np.sum(sensitive == 0)
    rate_y1_s1 = np.sum(labels[sensitive == 1] == 1) / np.sum(sensitive == 1)
    return rate_y1_s0, rate_y1_s1


def compute_odds(labels, predictions, sensitive):
    """Compute the bias in the predictions with respect to the sensitive attr. and the labels"""
    return dict(
        p_ybary0_s0=np.mean(predictions[(labels == 0) & (sensitive == 0)] == 0),
        p_ybary1_s0=np.mean(predictions[(labels == 1) & (sensitive == 0)] == 1),
        p_ybary0_s1=np.mean(predictions[(labels == 0) & (sensitive == 1)] == 0),
        p_ybary1_s1=np.mean(predictions[(labels == 1) & (sensitive == 1)] == 1),
    )


def _fix_labels(labels):
    """Make sure that labels are either 0 or 1

    Args"
        labels: the labels as a list of numpy arrays
        positive_class_val: the value that corresponds to a "positive" predictions

    Returns:
        the fixed labels and a function to convert the fixed labels back to the original format
    """
    label_values = list(np.unique(labels[0]))
    if label_values == [0, 1]:

        def _do_nothing(inp):
            return inp

        return labels, _do_nothing
    if label_values == [-1, 1]:

        def _converter(label):
            return (2 * label) - 1

        return [(y + 1) / 2 for y in labels], _converter
    raise ValueError("Labels have unknown structure")


def split_train_dev(inputs, labels, sensitive):
    """Split the given data into train and dev set with the proportion of labels being preserved"""
    n_total = inputs.shape[0]
    idx_s0_y0 = ((sensitive == 0) & (labels == 0)).nonzero()[0]
    idx_s0_y1 = ((sensitive == 0) & (labels == 1)).nonzero()[0]
    idx_s1_y0 = ((sensitive == 1) & (labels == 0)).nonzero()[0]
    idx_s1_y1 = ((sensitive == 1) & (labels == 1)).nonzero()[0]

    train_fraction: List[int] = []
    test_fraction: List[int] = []
    for idx in [idx_s0_y0, idx_s0_y1, idx_s1_y0, idx_s1_y1]:
        np.random.shuffle(idx)

        split_idx = int(len(idx) * 0.5) + 1  # make sure the train part is at least half
        train_fraction_a = idx[:split_idx]
        test_fraction_a = idx[split_idx:]
        train_fraction += list(train_fraction_a)
        test_fraction += list(test_fraction_a)
    xtrain, ytrain, strain = (
        inputs[train_fraction],
        labels[train_fraction],
        sensitive[train_fraction],
    )
    # ensure that the train set has exactly the same size as the given set
    # (otherwise inducing inputs has wrong shape)
    return dict(
        xtrain=np.concatenate((xtrain, xtrain))[:n_total],
        ytrain=np.concatenate((ytrain, ytrain))[:n_total],
        strain=np.concatenate((strain, strain))[:n_total],
        xtest=inputs[test_fraction],
        ytest=labels[test_fraction],
        stest=sensitive[test_fraction],
    )


def _flags(
    parameters, data_path, save_dir, s_as_input, model_name, num_train, gpu, epochs, length_scale
):
    batch_size = min(MAX_BATCH_SIZE, num_train)
    epochs = _num_epochs(num_train) if epochs is None else epochs
    return {
        **dict(
            inf="Variational",
            data="sensitive_from_numpy",
            dataset_path=data_path,
            cov="RBFKernel",
            mean="ZeroMean",
            optimizer="Adam",
            lr=0.05,
            # lr=0.1,
            model_name=model_name,
            batch_size=batch_size,
            # epochs=min(MAX_EPOCHS, _num_epochs(num_train)),
            epochs=min(epochs, MAX_EPOCHS),
            eval_epochs=100000,  # we can unfortunately not run evaluations because of no y_test
            summary_steps=100000,
            chkpt_epochs=100000,
            save_dir=save_dir,  # "/home/ubuntu/out2/",
            plot="",
            logging_steps=1,
            gpus=str(gpu),
            preds_path=PRED_FNAME,  # save the predictions into `predictions.npz`
            num_samples=1000,
            optimize_inducing=True,
            length_scale=length_scale,
            sf=1.0,
            iso=False,
            num_samples_pred=2000,
            s_as_input=s_as_input,
            # num_inducing=MAX_NUM_INDUCING,
            num_inducing=_num_inducing(num_train),
            manual_seed=SEED,
            # metrics=(
            #     "binary_accuracy,pred_rate_y1_s0,pred_rate_y1_s1,base_rate_y1_s0,base_rate_y1_s1,"
            #     "pred_odds_yhaty1_s0,pred_odds_yhaty1_s1,pred_odds_yhaty0_s0,pred_odds_yhaty0_s1"
            # ),
        ),
        **parameters,
    }


def _num_inducing(num_train):
    """Adaptive number of inducing inputs

    num_train == 4,000 => num_inducing == 1121
    num_train == 20,000 => num_inducing == 2507
    """
    return int(2500 / 141 * np.sqrt(num_train))


def _num_epochs(num_train):
    """Adaptive number of epochs

    num_train == 4,000 => num_epochs == 125.7
    num_train == 20,000 => num_epochs == 84
    """
    return int(1000 / np.power(num_train, 1 / 4))



"""
implementation of Agarwal model
"""
from pathlib import Path
from typing import List

import pandas as pd

from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmAsync
from ethicml.algorithms.inprocess.interface import conventional_interface
from ethicml.utility.data_structures import PathTuple, TestPathTuple, DataTuple, TestTuple
from ethicml.implementations import agarwal


VALID_FAIRNESS = {"DP", "EqOd"}
VALID_MODELS = {"LR", "SVM"}


class Agarwal(InAlgorithmAsync):
    """
    Agarwal class
    """

    def __init__(
        self,
        fairness: str = "DP",
        classifier: str = "LR",
        eps: float = 0.1,
        iters=50,
        C=None,
        kernel=None,
    ):
        if fairness not in VALID_FAIRNESS:
            raise ValueError("results: fairness must be one of %r." % VALID_FAIRNESS)
        if classifier not in VALID_MODELS:
            raise ValueError("results: classifier must be one of %r." % VALID_MODELS)
        super().__init__()
        self.classifier = classifier
        self.fairness = fairness
        self.eps = eps
        self.iters = iters
        self.C = C
        if self.C is None:
            if self.classifier == "LR":
                self.C = LogisticRegression().C
            elif self.classifier == "SVM":
                self.C = SVC().C
            else:
                raise NotImplementedError("Sever problem: this point should not be reached")
        self.kernel = kernel
        if self.kernel is None:
            if self.classifier == "SVM":
                self.kernel = SVC().kernel
            else:
                self.kernel = ""

    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        return agarwal.train_and_predict(
            train=train,
            test=test,
            classifier=self.classifier,
            fairness=self.fairness,
            eps=self.eps,
            iters=self.iters,
            C=self.C,
            kernel=self.kernel,
        )

    def _script_command(
        self, train_paths: PathTuple, test_paths: TestPathTuple, pred_path: Path
    ) -> (List[str]):
        script = ["-m", agarwal.train_and_predict.__module__]
        args = conventional_interface(
            train_paths,
            test_paths,
            pred_path,
            str(self.classifier),
            str(self.fairness),
            str(self.eps),
            str(self.iters),
            str(self.C),
            self.kernel,
        )
        return script + args

    @property
    def name(self) -> str:
        return f"Agarwal {self.classifier}"



"""
Collection of functions that enable parallelism
"""
import os
import asyncio
from typing import List, Tuple, Dict

import pandas as pd

from ethicml.utility.data_structures import DataTuple
from ethicml.algorithms.inprocess import InAlgorithm, InAlgorithmAsync


Data = Tuple[DataTuple, DataTuple]


async def arrange_in_parallel(
    algos: List[InAlgorithmAsync], data: List[Data], max_parallel: int = 0
) -> List[List[pd.DataFrame]]:
    """Arrange the given algorithms to run (embarrassingly) parallel

    Args:
        algos: list of algorithms that implement the `run_async` function
        data: list of pairs of data tuples (train and test)
        max_parallel: how many processes can run in parallel at most. if zero (or negative), then
                      there is no maximum
    Returns:
        list of the results
    """
    assert len(algos) >= 1
    assert len(data) >= 1
    assert isinstance(data[0][0], DataTuple)
    assert isinstance(data[0][1], DataTuple)
    # create queue of tasks
    task_queue: asyncio.Queue = asyncio.Queue()
    # for each algorithm, first loop over all available datasets and then go on to the next algo
    for i, algo in enumerate(algos):
        for j, data_item in enumerate(data):
            task_queue.put_nowait((i, j, algo, data_item))
    # create workers
    num_cpus = os.cpu_count()
    default_num_workers: int = num_cpus if num_cpus is not None else 1
    num_workers = max_parallel if max_parallel > 0 else default_num_workers
    result_dict: Dict[Tuple[int, int], pd.DataFrame] = {}
    workers = [_eval_worker(worker_id, task_queue, result_dict) for worker_id in range(num_workers)]
    # run workers and confirm that the queue is empty
    await asyncio.gather(*workers)
    assert task_queue.empty()
    # turn dictionary into a list
    return [[result_dict[(i, j)] for j in range(len(data))] for i in range(len(algos))]


async def _eval_worker(
    worker_id: int,
    task_queue: "asyncio.Queue[Tuple[int, int, InAlgorithmAsync, Data]]",
    result_dict: Dict[Tuple[int, int], pd.DataFrame],
) -> None:
    while not task_queue.empty():
        # get a work item out of the queue
        algo_id, data_id, algo, (train, test) = task_queue.get_nowait()
        # do the work
        result = await algo.run_async(train, test)
        # put result into results dictionary
        result_dict[(algo_id, data_id)] = result
        print(f"Worker {worker_id} has finished algo {algo_id} on dataset {data_id}.")
        # notify the queue that the work item has been processed
        task_queue.task_done()


def run_in_parallel(
    algos: List[InAlgorithm], data: List[Data], max_parallel: int = -1
) -> List[List[pd.DataFrame]]:
    """Run the given algorithms (embarrassingly) parallel

    Args:
        algos: list of in-process algorithms
        data: list of pairs of data tuples (train and test)
        max_parallel: how many processes can run in parallel at most. if negative, then there is no
                      maximum
    Returns:
        list of the results
    """
    # Filter out those algorithms that actually can run in their own process.
    # This is unfortunately really complicated because we have to keep track of the indices
    # in order to reassemble the returned list correctly.
    async_algos: List[InAlgorithmAsync] = []
    async_idx: List[int] = []
    blocking_algos: List[InAlgorithm] = []
    blocking_idx: List[int] = []
    for i, algo in enumerate(algos):
        if isinstance(algo, InAlgorithmAsync):
            async_algos.append(algo)
            async_idx.append(i)
        else:
            blocking_algos.append(algo)
            blocking_idx.append(i)
    event_loop = asyncio.get_event_loop()

    # first start the asynchronous results
    async_coroutines = arrange_in_parallel(async_algos, data, max_parallel)

    # then get the blocking results
    # for each algorithm, first loop over all available datasets and then go on to the next algo
    blocking_results = [[algo.run(train, test) for train, test in data] for algo in blocking_algos]

    # then wait for the asynchronous results to come in
    async_results = event_loop.run_until_complete(async_coroutines)

    # now reassemble everything in the right order
    results: List[List[pd.DataFrame]] = []
    async_counter = 0
    blocking_counter = 0
    for i in range(len(algos)):
        if i in async_idx:
            results.append(async_results[async_counter])
            async_counter += 1
        else:
            results.append(blocking_results[blocking_counter])
            blocking_counter += 1
    return results



"""Pre-process algorithms take the training data and transform it"""
from .beutel import Beutel
from .pre_algorithm import PreAlgorithmAsync, PreAlgorithm
from .zemel import Zemel
from .vfae import VFAE



"""
Useful functions used in implementations
"""
import sys
import argparse
from pathlib import Path
from typing import Tuple, Union, List, Optional, Dict, Any

import pandas as pd
import numpy as np

from ethicml.utility.data_structures import DataTuple, TestTuple
from ethicml.algorithms.algorithm_base import load_dataframe


class InAlgoInterface:
    """Commandline "interface" for in-process algorithms"""

    def __init__(self, args: Optional[List[str]] = None):
        self.args: List[str] = sys.argv[1:] if args is None else args

    def load_data(self) -> Tuple[DataTuple, TestTuple]:
        """Load the data from the files"""
        train = DataTuple(
            x=load_dataframe(Path(self.args[0])),
            s=load_dataframe(Path(self.args[1])),
            y=load_dataframe(Path(self.args[2])),
            name=str(load_dataframe(Path(self.args[3]))['0'][0]),
        )
        test = TestTuple(
            x=load_dataframe(Path(self.args[4])),
            s=load_dataframe(Path(self.args[5])),
            name=str(load_dataframe(Path(self.args[6]))['0'][0]),
        )
        return train, test

    def save_predictions(self, predictions: Union[np.ndarray, pd.DataFrame]):
        """Save the data to the file that was specified in the commandline arguments"""
        if not isinstance(predictions, pd.DataFrame):
            df = pd.DataFrame(predictions, columns=["pred"])
        else:
            df = predictions
        pred_path = Path(self.args[7])
        df.to_feather(pred_path)

    def remaining_args(self) -> List[str]:
        """Additional commandline arguments beyond the data paths and the prediction path"""
        return self.args[8:]


def load_data_from_flags(flags: Dict[str, Any]) -> Tuple[DataTuple, TestTuple]:
    """Load data from the paths specified in the flags"""
    train = DataTuple(
        x=load_dataframe(Path(flags["train_x"])),
        s=load_dataframe(Path(flags["train_s"])),
        y=load_dataframe(Path(flags["train_y"])),
        name=str(load_dataframe(Path(flags["train_name"]))['0'][0]),
    )
    test = TestTuple(
        x=load_dataframe(Path(flags['test_x'])),
        s=load_dataframe(Path(flags['test_s'])),
        name=str(load_dataframe(Path(flags["test_name"]))['0'][0]),
    )
    return train, test


def save_transformations(transforms: Tuple[DataTuple, TestTuple], args: argparse.Namespace):
    """Save the data to the file that was specified in the commandline arguments"""
    assert isinstance(transforms[0], DataTuple)
    assert isinstance(transforms[1], TestTuple)

    train, test = transforms

    def _save(data: pd.DataFrame, path: str) -> None:
        # SUGGESTION: maybe the file names should be completely random to avoid collisions
        data_path = Path(path)
        # write the file
        data.columns = data.columns.astype(str)
        data.to_feather(data_path)

    _save(train.x, args.new_train_x)
    _save(train.s, args.new_train_s)
    _save(train.y, args.new_train_y)
    _save(pd.DataFrame([train.name], columns=['0']), args.new_train_name)
    _save(test.x, args.new_test_x)
    _save(test.s, args.new_test_s)
    _save(pd.DataFrame([test.name], columns=['0']), args.new_test_name)


def pre_algo_argparser() -> argparse.ArgumentParser:
    """ArgumentParser that already has arguments for the paths

    This is a quick way to create a parser that can parse the filenames for the data. This function
    can be used by pre-algorithms to implement a commandline interface.
    """
    parser = argparse.ArgumentParser()

    # paths to the files with the data
    parser.add_argument("--train_x", required=True)
    parser.add_argument("--train_s", required=True)
    parser.add_argument("--train_y", required=True)
    parser.add_argument("--train_name", required=True)
    parser.add_argument("--test_x", required=True)
    parser.add_argument("--test_s", required=True)
    parser.add_argument("--test_name", required=True)

    # paths to where the processed inputs should be stored
    parser.add_argument("--new_train_x", required=True)
    parser.add_argument("--new_train_s", required=True)
    parser.add_argument("--new_train_y", required=True)
    parser.add_argument("--new_train_name", required=True)
    parser.add_argument("--new_test_x", required=True)
    parser.add_argument("--new_test_s", required=True)
    parser.add_argument("--new_test_name", required=True)
    return parser



"""
Implementation for Louizos et al Variational Fair Autoencoder
"""

import torch
import torch.nn.functional as F

from ethicml.implementations.pytorch_common import quadratic_time_mmd


def kullback_leibler(mu1, logvar1, mu2=None, logvar2=None):
    """

    Args:
        mu1:
        logvar1:
        mu2:
        logvar2:

    Returns:

    """
    mu2 = mu2 if mu2 is not None else torch.tensor([0.0])
    logvar2 = logvar2 if logvar2 is not None else torch.tensor([0.0])
    # return -0.5 * torch.sum(1 + (logvar1-logvar2) - (mu1-mu2).pow(2) - (logvar1-logvar2).exp())
    return (
        0.5
        * (logvar2 - logvar1 - 1 + ((logvar1).exp() + ((mu1 - mu2) ** 2)) / (logvar2).exp()).sum()
    )


def loss_function(flags, z1_triplet, z2_triplet, z1_d_triplet, data_triplet, x_dec, y_pred):
    """

    Args:
        flags:
        z1_triplet:
        z2_triplet:
        z1_d_triplet:
        data_triplet:
        x_dec:
        y_pred:

    Returns:

    """
    z1, z1_mu, z1_logvar = z1_triplet
    if flags["supervised"]:
        z2, z2_mu, z2_logvar = z2_triplet
        z1_dec, z1_dec_mu, z1_dec_logvar = z1_d_triplet
    x, s, y = data_triplet

    reconstruction_loss = F.mse_loss(x_dec, x, reduction="sum")

    if flags["fairness"] == "DI":
        z1_s0 = torch.masked_select(z1, s.le(0.5)).view(-1, 50)
        z1_s1 = torch.masked_select(z1, s.ge(0.5)).view(-1, 50)
    elif flags["fairness"] == "Eq. Opp":
        z1_s0 = torch.masked_select(z1, s.le(0.5)).view(-1, 50)
        y_s0 = torch.masked_select(y, s.le(0.5)).view(-1, 1)
        z1_s0_y1 = torch.masked_select(z1_s0, y_s0.ge(0.5)).view(-1, 50)
        z1_s1 = torch.masked_select(z1, s.ge(0.5)).view(-1, 50)
        y_s1 = torch.masked_select(y, s.ge(0.5)).view(-1, 1)
        z1_s1_y1 = torch.masked_select(z1_s1, y_s1.ge(0.5)).view(-1, 50)

        z1_s0 = z1_s0_y1
        z1_s1 = z1_s1_y1
    else:
        raise NotImplementedError("Only DI and Eq.Opp implementesd so far")

    mmd_loss = quadratic_time_mmd(z1_s0, z1_s1, 2.5)

    if flags["supervised"]:
        first_kl = kullback_leibler(z2_mu, z2_logvar)
        second_kl = kullback_leibler(z1_dec_mu, z1_dec_logvar, z1_mu, z1_logvar)
        # second_kl = F.kl_div(z1_dec, z1, reduction='sum')
        # second_kl = (z1_dec.sum()+1e-10).log() - (z1.sum()+1e-10).log()
        kl_div = first_kl + second_kl
        prediction_loss = F.binary_cross_entropy(y_pred, y, reduction="sum")
    else:
        kl_div = 0
        prediction_loss = 0

    return prediction_loss, reconstruction_loss, kl_div, 100 * mmd_loss



"""Implementation of logistic regression (actually just a wrapper around sklearn)"""
from sklearn.linear_model import LogisticRegressionCV
from sklearn.model_selection import KFold

import pandas as pd

from ethicml.implementations.utils import InAlgoInterface


def train_and_predict(train, test):
    """Train a logistic regression model and compute predictions on the given test data"""
    folder = KFold(n_splits=3, random_state=888, shuffle=False)
    clf = LogisticRegressionCV(
        cv=folder, n_jobs=-1, random_state=888, solver="liblinear", multi_class='auto'
    )
    clf.fit(train.x, train.y.values.ravel())
    return pd.DataFrame(clf.predict(test.x), columns=["preds"])


def main():
    """This function runs the cross-validated logistic regression model as a standalone program"""
    interface = InAlgoInterface()
    train, test = interface.load_data()
    interface.save_predictions(train_and_predict(train, test))


if __name__ == "__main__":
    main()



"""
Test that an algorithm can run against some data
"""

from typing import Tuple
import numpy as np

from ethicml.algorithms.inprocess import LR, SVM, Majority
from ethicml.utility.data_structures import DataTuple
from ethicml.data.load import load_data
from ethicml.data import Toy
from ethicml.evaluators import run_in_parallel
from ethicml.preprocessing.train_test_split import train_test_split


def get_train_test() -> Tuple[DataTuple, DataTuple]:
    data: DataTuple = load_data(Toy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    return train_test


def count_true(mask: 'np.ndarray[bool]') -> int:
    """Count the number of elements that are True"""
    return mask.nonzero()[0].shape[0]


def test_can_load_test_data():
    train, test = get_train_test()
    assert train is not None
    assert test is not None


def test_run_parallel():
    data0 = get_train_test()
    data1 = get_train_test()
    result = run_in_parallel([LR(), SVM(), Majority()], [data0, data1], max_parallel=2)
    # LR
    assert count_true(result[0][0].values == 1) == 211
    assert count_true(result[0][0].values == -1) == 189
    assert count_true(result[0][1].values == 1) == 211
    assert count_true(result[0][1].values == -1) == 189
    # SVM
    assert count_true(result[1][0].values == 1) == 201
    assert count_true(result[1][0].values == -1) == 199
    assert count_true(result[1][1].values == 1) == 201
    assert count_true(result[1][1].values == -1) == 199
    # Majority
    assert count_true(result[2][0].values == 1) == 0
    assert count_true(result[2][0].values == -1) == 400
    assert count_true(result[2][1].values == 1) == 0
    assert count_true(result[2][1].values == -1) == 400


# def test_run_alg_suite():
#     dataset = Adult("Race")
#     dataset.sens_attrs = ["race_White"]
#     datasets: List[Dataset] = [dataset, Toy()]
#     preprocess_models: List[PreAlgorithm] = [Beutel()]
#     inprocess_models: List[InAlgorithm] = [LR()]
#     postprocess_models: List[PostAlgorithm] = []
#     metrics: List[Metric] = [Accuracy(), CV()]
#     per_sens_metrics: List[Metric] = [Accuracy(), TPR()]
#     evaluate_models(datasets, preprocess_models, inprocess_models,
#                     postprocess_models, metrics, per_sens_metrics, repeats=1, test_mode=True)
#
#
# def test_run_alg_suite_wrong_metrics():
#     datasets: List[Dataset] = [Toy(), Adult()]
#     preprocess_models: List[PreAlgorithm] = [Beutel()]
#     inprocess_models: List[InAlgorithm] = [SVM(), LR()]
#     postprocess_models: List[PostAlgorithm] = []
#     metrics: List[Metric] = [Accuracy(), CV()]
#     per_sens_metrics: List[Metric] = [Accuracy(), TPR(), CV()]
#     with pytest.raises(MetricNotApplicable):
#         evaluate_models(datasets, preprocess_models, inprocess_models,
#                         postprocess_models, metrics, per_sens_metrics, repeats=1, test_mode=True)



"""In-process algorithms take training data and make predictions"""
from .agarwal_reductions import Agarwal
from .in_algorithm import InAlgorithm, InAlgorithmAsync
from .fair_gpyt import GPyT, GPyTDemPar, GPyTEqOdds
from .kamishima import Kamishima
from .kamiran import Kamiran
from .logistic_regression import LR, LRCV, LRProb
from .svm import SVM
from .majority import Majority
from .mlp import MLP



"""
Zemel's Learned Fair Representations
"""
from pathlib import Path
from typing import Dict, Union, Tuple, List

from ethicml.utility.data_structures import TestTuple, DataTuple, PathTuple, TestPathTuple
from .pre_algorithm import PreAlgorithmAsync
from .interface import flag_interface


class Zemel(PreAlgorithmAsync):
    """
    AIF360 implementation of Zemel's LFR
    """

    def __init__(
        self,
        threshold=0.5,
        clusters=2,
        Ax=0.01,
        Ay=0.1,
        Az=0.5,
        max_iter=5000,
        maxfun=5000,
        epsilon=1e-5,
    ):
        super().__init__()
        self.flags: Dict[str, Union[int, float]] = {
            "clusters": clusters,
            "Ax": Ax,
            "Ay": Ay,
            "Az": Az,
            "max_iter": max_iter,
            "maxfun": maxfun,
            "epsilon": epsilon,
            "threshold": threshold,
        }

    def run(self, train, test) -> Tuple[DataTuple, TestTuple]:
        from ...implementations import zemel

        return zemel.train_and_transform(train, test, self.flags)

    def _script_command(
        self,
        train_paths: PathTuple,
        test_paths: TestPathTuple,
        new_train_x_path: Path,
        new_train_s_path: Path,
        new_train_y_path: Path,
        new_train_name_path: Path,
        new_test_x_path: Path,
        new_test_s_path: Path,
        new_test_name_path: Path,
    ) -> List[str]:
        args = flag_interface(
            train_paths,
            test_paths,
            new_train_x_path,
            new_train_s_path,
            new_train_y_path,
            new_train_name_path,
            new_test_x_path,
            new_test_s_path,
            new_test_name_path,
            self.flags,
        )
        return ["-m", "ethicml.implementations.zemel"] + args

    @property
    def name(self) -> str:
        return "Zemel"



"""
Abstract Base Class of all post-processing algorithms in the framework
"""

from abc import abstractmethod
import pandas as pd

from ethicml.utility.data_structures import DataTuple
from ..algorithm_base import Algorithm


class PostAlgorithm(Algorithm):
    """Abstract Base Class for all algorithms that do post-processing"""

    @abstractmethod
    def run(self, predictions: pd.DataFrame, test: DataTuple) -> pd.DataFrame:
        """Make predictions fair

        Args:
            predictions:
            test:
        """
        raise NotImplementedError("Run needs to be implemented")

    def run_test(self, predictions: pd.DataFrame, test: DataTuple) -> pd.DataFrame:
        """

        Args:
            predictions:
            test:
        """
        raise NotImplementedError("Need to reduce data to 500 datapoints")



"""
Wrapper around Sci-Kit Learn Logistic Regression
"""
from typing import Optional

from sklearn.linear_model import LogisticRegression

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmAsync
from ethicml.algorithms.inprocess.interface import conventional_interface
from ethicml.implementations import (
    logistic_regression,
    logistic_regression_probability,
    logistic_regression_cross_validated,
)


class LR(InAlgorithmAsync):
    """Logistic regression with hard predictions"""

    def __init__(self, C: Optional[float] = None):
        super().__init__()
        self.C = LogisticRegression().C if C is None else C

    def run(self, train, test):
        return logistic_regression.train_and_predict(train, test, self.C)

    def _script_command(self, train_paths, test_paths, pred_path):
        script = ["-m", logistic_regression.train_and_predict.__module__]
        args = conventional_interface(train_paths, test_paths, pred_path, str(self.C))
        return script + args

    @property
    def name(self) -> str:
        return "Logistic Regression"


class LRProb(InAlgorithmAsync):
    """Logistic regression with soft output"""

    def __init__(self, C: Optional[int] = None):
        super().__init__()
        self.C = LogisticRegression().C if C is None else C

    def run(self, train, test):
        return logistic_regression_probability.train_and_predict(train, test, self.C)

    def _script_command(self, train_paths, test_paths, pred_path):
        script = ["-m", logistic_regression_probability.train_and_predict.__module__]
        args = conventional_interface(train_paths, test_paths, pred_path, str(self.C))
        return script + args

    @property
    def name(self) -> str:
        return "Logistic Regression Prob"


class LRCV(InAlgorithmAsync):
    """Kind of a cheap hack for now, but gives a proper cross-valudeted LR"""

    def run(self, train, test):
        return logistic_regression_cross_validated.train_and_predict(train, test)

    def _script_command(self, train_paths, test_paths, pred_path):
        script = ["-m", logistic_regression_cross_validated.train_and_predict.__module__]
        args = conventional_interface(train_paths, test_paths, pred_path)
        return script + args

    @property
    def name(self) -> str:
        return "LRCV"



"""Methods that define commandline interfaces"""
from typing import List

from ethicml.utility.data_structures import PathTuple, TestPathTuple


def conventional_interface(train_paths: PathTuple, test_paths: TestPathTuple, *args) -> List[str]:
    """
    Generate the commandline arguments that are expected by the scripts that follow the convention.

    The agreed upon order is:
    x (train), s (train), y (train), x (test), s (test), predictions.
    """
    list_to_return = [
        str(train_paths.x),
        str(train_paths.s),
        str(train_paths.y),
        str(train_paths.name),
        str(test_paths.x),
        str(test_paths.s),
        str(test_paths.name),
    ]
    for arg in args:
        list_to_return.append(str(arg))
    return list_to_return



"""Kamiran and Calders 2012"""
from pathlib import Path
from typing import List, Optional

import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmAsync
from ethicml.algorithms.inprocess.interface import conventional_interface
from ethicml.utility.data_structures import PathTuple, TestPathTuple, DataTuple, TestTuple
from ethicml.implementations import kamiran

VALID_MODELS = {"LR", "SVM"}


class Kamiran(InAlgorithmAsync):
    """
    Kamiran and Calders 2012
    """

    def __init__(
        self, classifier: str = "LR", C: Optional[float] = None, kernel: Optional[str] = None
    ):
        super().__init__()
        if classifier not in VALID_MODELS:
            raise ValueError("results: classifier must be one of %r." % VALID_MODELS)
        self.classifier = classifier
        if C is None:
            if self.classifier == "LR":
                self.C = LogisticRegression().C
            elif self.classifier == "SVM":
                self.C = SVC().C
            else:
                raise NotImplementedError("Sever problem: this point should not be reached")
        else:
            self.C = C
        if kernel is None:
            if self.classifier == "SVM":
                self.kernel = SVC().kernel
            else:
                self.kernel = ""
        else:
            self.kernel = kernel

    def run(self, train: DataTuple, test: TestTuple) -> pd.DataFrame:
        return kamiran.train_and_predict(
            train, test, classifier=self.classifier, C=self.C, kernel=self.kernel
        )

    def _script_command(
        self, train_paths: PathTuple, test_paths: TestPathTuple, pred_path: Path
    ) -> (List[str]):
        script = ["-m", kamiran.train_and_predict.__module__]
        args = conventional_interface(
            train_paths, test_paths, pred_path, self.classifier, str(self.C), self.kernel
        )
        return script + args

    @property
    def name(self) -> str:
        return f"Kamiran & Calders {self.classifier}"



"""
For assessing Calder-Verwer metric: 1-(P(Y=1|S=1)-P(Y=1|S!=1))
"""

import pandas as pd

from ethicml.metrics.prob_pos import ProbPos
from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class CV(Metric):
    """Calder-Verwer"""

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        from ethicml.evaluators.per_sensitive_attribute import (
            metric_per_sensitive_attribute,
            diff_per_sensitive_attribute,
        )

        per_sens = metric_per_sensitive_attribute(prediction, actual, ProbPos())
        diffs = diff_per_sensitive_attribute(per_sens)

        return 1 - list(diffs.values())[0]

    @property
    def name(self) -> str:
        return "CV"

    @property
    def apply_per_sensitive(self) -> bool:
        return False



"""
For assessing TNR
"""

import pandas as pd

from ethicml.utility.data_structures import DataTuple
from .confusion_matrix import confusion_matrix
from .metric import Metric


class TNR(Metric):
    """True negative rate"""

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        t_neg, f_pos, _, _ = confusion_matrix(prediction, actual, pos_cls=self.positive_class)

        return t_neg / (t_neg + f_pos)

    @property
    def name(self) -> str:
        return "TNR"



"""
For assessing ProbPos
"""

import pandas as pd

from ethicml.utility.data_structures import DataTuple
from .confusion_matrix import confusion_matrix
from .metric import Metric


class ProbPos(Metric):
    """Probability of positive prediction"""

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        _, f_pos, _, t_pos = confusion_matrix(prediction, actual, pos_cls=self.positive_class)

        return (t_pos + f_pos) / prediction.size

    @property
    def name(self) -> str:
        return "prob_pos"



"""
Runs given metrics on given algorithms for given datasets
"""
import os
from pathlib import Path
from typing import List, Dict, Union

import pandas as pd
from tqdm import tqdm

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.algorithms.postprocess.post_algorithm import PostAlgorithm
from ethicml.algorithms.preprocess.pre_algorithm import PreAlgorithm
from ethicml.utility.data_structures import DataTuple, get_subset, TestTuple
from ..data.dataset import Dataset
from ..data.load import load_data
from .per_sensitive_attribute import (
    metric_per_sensitive_attribute,
    MetricNotApplicable,
    diff_per_sensitive_attribute,
    ratio_per_sensitive_attribute,
)
from ..metrics.metric import Metric
from ..preprocessing.train_test_split import train_test_split


def get_sensitive_combinations(metrics: List[Metric], train: DataTuple) -> List[str]:
    """Get all possible combinations of sensitive attribute and metrics"""
    poss_values = []
    for col in train.s.columns:
        uniques = train.s[col].unique()
        for unique in uniques:
            poss_values.append(f"{col}_{unique}")

    return [f"{s}_{m.name}" for s in poss_values for m in metrics]


def per_sens_metrics_check(per_sens_metrics: List[Metric]):
    """Check if the given metrics allow application per sensitive attribute"""
    for metric in per_sens_metrics:
        if not metric.apply_per_sensitive:
            raise MetricNotApplicable(
                f"Metric {metric.name} is not applicable per sensitive "
                f"attribute, apply to whole dataset instead"
            )


def run_metrics(
    predictions: pd.DataFrame,
    actual: DataTuple,
    metrics: List[Metric],
    per_sens_metrics: List[Metric],
) -> Dict[str, float]:
    """Run all the given metrics on the given predictions and return the results

    Args:
        predictions: DataFrame with predictions
        actual: DataTuple with the labels
        metrics: list of metrics
        per_sens_metrics: list of metrics that are computed per sensitive attribute
    """
    result: Dict[str, float] = {}
    for metric in metrics:
        result[metric.name] = metric.score(predictions, actual)

    for metric in per_sens_metrics:
        per_sens = metric_per_sensitive_attribute(predictions, actual, metric)
        diff_per_sens = diff_per_sensitive_attribute(per_sens)
        ratio_per_sens = ratio_per_sensitive_attribute(per_sens)
        per_sens.update(diff_per_sens)
        per_sens.update(ratio_per_sens)
        for key, value in per_sens.items():
            result[f"{metric.name}_{key}"] = value
    return result  # SUGGESTION: we could return a DataFrame here instead of a dictionary


def evaluate_models(
    datasets: List[Dataset],
    preprocess_models: List[PreAlgorithm],
    inprocess_models: List[InAlgorithm],
    postprocess_models: List[PostAlgorithm],
    metrics: List[Metric],
    per_sens_metrics: List[Metric],
    repeats: int = 3,
    test_mode: bool = False,
) -> pd.DataFrame:
    """Evaluate all the given models for all the given datasets and compute all the given metrics

    Args:
        datasets: list of dataset objects
        preprocess_models: list of preprocess model objects
        inprocess_models: list of inprocess model objects
        postprocess_models: list of postprocess model objects
        metrics: list of metric objects
        per_sens_metrics: list of metric objects that will be evaluated per sensitive attribute
        test_mode: if True, only use a small subset of the data so that the models run faster
    """
    per_sens_metrics_check(per_sens_metrics)

    columns = ["dataset", "transform", "model", "repeat"]
    columns += [metric.name for metric in metrics]
    results = pd.DataFrame(columns=columns)

    total_experiments = (
        len(datasets)
        * repeats
        * (1 + len(preprocess_models) + ((1 + len(preprocess_models)) * len(inprocess_models)))
    )

    seed = 0
    with tqdm(total=total_experiments) as pbar:
        for dataset in datasets:
            for repeat in range(repeats):
                train: DataTuple
                test: DataTuple
                train, test = train_test_split(load_data(dataset), random_seed=seed)
                seed += 2410
                if test_mode:
                    # take smaller subset of training data to speed up training
                    train = get_subset(train)

                to_operate_on: Dict[str, Dict[str, Union[DataTuple, TestTuple]]] = {
                    "no_transform": {"train": train, "test": test}
                }

                for pre_process_method in preprocess_models:
                    new_train, new_test = pre_process_method.run(train, test)
                    to_operate_on[pre_process_method.name] = {"train": new_train, "test": new_test}

                    pbar.update()

                transform_name: str
                for transform_name, transform in to_operate_on.items():

                    transformed_train: DataTuple = transform["train"]
                    transformed_test: DataTuple = transform["test"]

                    for model in inprocess_models:

                        temp_res: Dict[str, Union[str, float]] = {
                            "dataset": dataset.name,
                            "transform": transform_name,
                            "model": model.name,
                            "repeat": f"{repeat}-{seed}",
                        }

                        predictions: pd.DataFrame
                        predictions = model.run(transformed_train, transformed_test)

                        temp_res.update(run_metrics(predictions, test, metrics, per_sens_metrics))

                        for postprocess in postprocess_models:
                            # Post-processing has yet to be defined
                            # - leaving blank until we have an implementation to work with
                            pass

                        pbar.update()

                        results = results.append(temp_res, ignore_index=True)
                    outdir = Path("..") / "results"  # OS-independent way of saying '../results'
                    outdir.mkdir(exist_ok=True)
                    path_to_file = outdir / f"{dataset.name}_{transform_name}.csv"
                    exists = os.path.isfile(path_to_file)
                    if exists:
                        loaded_results = pd.read_csv(path_to_file)
                        results = pd.concat([loaded_results, results])
                    results.to_csv(path_to_file, index=False)

                    pbar.update()

    results = results.set_index(["dataset", "transform", "model", "repeat"])
    return results



"""
Useful methods that are used in some of the data objects
"""

from typing import List


def get_concatenated_features(
    s_prefix: List[str], y_prefix: List[str], continuous_features: List[str], discrete_only: bool
) -> List[str]:
    """

    Args:
        s_prefix:
        y_prefix:
        continuous_features:
        discrete_only:

    Returns:

    """
    return s_prefix + y_prefix + continuous_features if discrete_only else s_prefix + y_prefix


def filter_features_by_prefixes(features: List[str], prefixes: List[str]):
    """Filter the features by prefixes

    Args:
        features: list of features names
        prefixes: list of prefixes

    Returns:
        filtered feature names
    """
    res = []
    for name in features:
        filtered = False
        for pref in prefixes:
            if name.startswith(pref):
                filtered = True
                break
        if not filtered:
            res.append(name)
    return res


def get_discrete_features(
    all_feats: List[str], feats_to_remove: List[str], cont_feats: List[str]
) -> List[str]:
    """

    Args:
        all_feats:
        feats_to_remove:
        cont_feats:

    Returns:

    """
    return [
        item
        for item in filter_features_by_prefixes(all_feats, feats_to_remove)
        if item not in cont_feats
    ]



"""
For assessing mean of logits
"""

import pandas as pd

from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class ProbOutcome(Metric):
    """Mean of logits"""

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        return prediction.values.sum() / prediction.size

    @property
    def name(self) -> str:
        return "prob_outcome"



"""
For assessing Theil Index metric: based on aif 360 implementation
https://github.com/IBM/AIF360/blob/b5f4591c7c7374b9c88706827080a931209e7a24/aif360/metrics/classification_metric.py
in turn based on the paper https://arxiv.org/abs/1807.00787
"""

import numpy as np
import pandas as pd

from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class Theil(Metric):
    """Theil Index"""

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        y_true_df = actual.y
        act_col = y_true_df.columns[0]
        y_pos_label = y_true_df[act_col].max()

        y_pred = prediction["preds"].values.ravel()
        y_true = y_true_df.values.ravel()
        y_pred = (y_pred == y_pos_label).astype(np.float64)
        y_true = (y_true == y_pos_label).astype(np.float64)
        var_b = 1 + y_pred - y_true

        # moving the b inside the log allows for 0 values
        return float(np.mean(np.log((var_b / np.mean(var_b)) ** var_b) / np.mean(var_b)))

    @property
    def name(self) -> str:
        return "Theil_Index"



"""
Testing for plotting functionality
"""
from typing import Tuple

from ethicml.utility.data_structures import DataTuple
from ethicml.data import Adult
from ethicml.data.load import load_data
from ethicml.preprocessing.train_test_split import train_test_split
from ethicml.visualisation.plot import save_2d_plot, save_label_plot, save_jointplot
from tests.run_algorithm_test import get_train_test


def test_plot():
    train, _ = get_train_test()
    save_2d_plot(train, "plots/test.png")


def test_joint_plot():
    train, _ = get_train_test()
    save_jointplot(train, "plots/joint.png")


def test_label_plot():
    data: DataTuple = load_data(Adult())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, _ = train_test

    # train, _ = get_train_test()
    save_label_plot(train, "plots/labels.png")



"""Methods that define commandline interfaces"""
from typing import List, Any, Dict
from dataclasses import asdict

from ethicml.utility.data_structures import PathTuple, TestPathTuple


def flag_interface(
    train_paths: PathTuple,
    test_paths: TestPathTuple,
    new_train_x_path,
    new_train_s_path,
    new_train_y_path,
    new_train_name_path,
    new_test_x_path,
    new_test_s_path,
    new_test_name_path,
    flags: Dict[str, Any],
) -> List[str]:
    """Generate the commandline arguments that are expected"""
    flags_list: List[str] = []

    # paths to training and test data
    path_tuples = [train_paths, test_paths]
    prefixes = ["--train_", "--test_"]
    for path_tuple, prefix in zip(path_tuples, prefixes):
        for key, path in asdict(path_tuple).items():
            flags_list += [f"{prefix}{key}", str(path)]

    # paths to output files
    flags_list += [
        "--new_train_x",
        str(new_train_x_path),
        "--new_train_s",
        str(new_train_s_path),
        "--new_train_y",
        str(new_train_y_path),
        "--new_train_name",
        str(new_train_name_path),
        "--new_test_x",
        str(new_test_x_path),
        "--new_test_s",
        str(new_test_s_path),
        "--new_test_name",
        str(new_test_name_path),
    ]

    # model parameters
    for key, values in flags.items():
        flags_list.append(f"--{key}")
        if isinstance(values, (list, tuple)):
            flags_list += [str(value) for value in values]
        else:
            flags_list.append(str(values))
    return flags_list



"""
For assessing TPR
"""

import pandas as pd

from ethicml.utility.data_structures import DataTuple
from .confusion_matrix import confusion_matrix
from .metric import Metric


class TPR(Metric):
    """True positive rate"""

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        _, _, f_neg, t_pos = confusion_matrix(prediction, actual, self.positive_class)

        return t_pos / (t_pos + f_neg)

    @property
    def name(self) -> str:
        return "TPR"



from typing import Tuple
import pandas as pd

from ethicml.algorithms.algorithm_base import run_blocking
from ethicml.algorithms.inprocess import InAlgorithm, SVM
from ethicml.algorithms.preprocess import PreAlgorithm, PreAlgorithmAsync, Beutel, Zemel
from ethicml.algorithms.preprocess.vfae import VFAE
from ethicml.utility.data_structures import DataTuple, FairType, TestTuple
from tests.run_algorithm_test import get_train_test


def test_beutel():
    train, test = get_train_test()

    beut_model: PreAlgorithm = Beutel()
    assert beut_model is not None
    assert beut_model.name == "Beutel"

    new_train_test: Tuple[DataTuple, TestTuple] = beut_model.run(train, test)
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    svm_model: InAlgorithm = SVM()
    assert svm_model is not None
    assert svm_model.name == "SVM"

    predictions: pd.DataFrame = svm_model.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 201
    assert predictions.values[predictions.values == -1].shape[0] == 199


def test_vfae():
    train, test = get_train_test()

    vfae_model: PreAlgorithm = VFAE(dataset="Toy", epochs=10, batch_size=100)
    assert vfae_model is not None
    assert vfae_model.name == "VFAE"

    new_train_test: Tuple[DataTuple, TestTuple] = vfae_model.run(train, test)
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    svm_model: InAlgorithm = SVM()
    assert svm_model is not None
    assert svm_model.name == "SVM"

    predictions: pd.DataFrame = svm_model.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 201
    assert predictions.values[predictions.values == -1].shape[0] == 199

    vfae_model = VFAE(dataset="Toy", epochs=10, fairness="Eq. Opp", batch_size=100)
    assert vfae_model is not None
    assert vfae_model.name == "VFAE"

    new_train_test = vfae_model.run(train, test)
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    predictions = svm_model.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 201
    assert predictions.values[predictions.values == -1].shape[0] == 199

    vfae_model = VFAE(dataset="Toy", supervised=False, epochs=10,
                      fairness="Eq. Opp", batch_size=100)
    assert vfae_model is not None
    assert vfae_model.name == "VFAE"

    new_train_test = vfae_model.run(train, test)
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    predictions = svm_model.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 207
    assert predictions.values[predictions.values == -1].shape[0] == 193


def test_zemel():
    train, test = get_train_test()

    zemel_model: PreAlgorithm = Zemel()
    assert zemel_model is not None
    assert zemel_model.name == "Zemel"

    new_train_test: Tuple[DataTuple, TestTuple] = zemel_model.run(train, test)
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    svm_model: InAlgorithm = SVM()
    assert svm_model is not None
    assert svm_model.name == "SVM"

    predictions: pd.DataFrame = svm_model.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 182
    assert predictions.values[predictions.values == -1].shape[0] == 218


def test_threaded_zemel():
    train, test = get_train_test()

    model: PreAlgorithmAsync = Zemel()
    assert model is not None
    assert model.name == "Zemel"

    new_train_test: Tuple[DataTuple, TestTuple] = run_blocking(
        model.run_async(train, test)
    )
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    classifier: InAlgorithm = SVM()
    assert classifier is not None
    assert classifier.name == "SVM"

    predictions: pd.DataFrame = classifier.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 182
    assert predictions.values[predictions.values == -1].shape[0] == 218

    beut_model: PreAlgorithm = Zemel()
    assert beut_model is not None
    assert beut_model.name == "Zemel"

    new_train_test = beut_model.run(train, test)
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    svm_model: InAlgorithm = SVM()
    assert svm_model is not None
    assert svm_model.name == "SVM"

    predictions = svm_model.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 182
    assert predictions.values[predictions.values == -1].shape[0] == 218


def test_threaded_beutel():
    train, test = get_train_test()

    model: PreAlgorithmAsync = Beutel()
    assert model is not None
    assert model.name == "Beutel"

    new_train_test: Tuple[DataTuple, TestTuple] = run_blocking(
        model.run_async(train, test)
    )
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    classifier: InAlgorithm = SVM()
    assert classifier is not None
    assert classifier.name == "SVM"

    predictions: pd.DataFrame = classifier.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 201
    assert predictions.values[predictions.values == -1].shape[0] == 199

    beut_model: PreAlgorithm = Beutel()
    assert beut_model is not None
    assert beut_model.name == "Beutel"

    new_train_test = beut_model.run(train, test)
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    svm_model: InAlgorithm = SVM()
    assert svm_model is not None
    assert svm_model.name == "SVM"

    predictions = svm_model.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 201
    assert predictions.values[predictions.values == -1].shape[0] == 199


def test_threaded_vfae():
    train, test = get_train_test()

    model: PreAlgorithmAsync = VFAE(dataset='Toy')
    assert model is not None
    assert model.name == "VFAE"

    new_train_test: Tuple[DataTuple, TestTuple] = run_blocking(
        model.run_async(train, test)
    )
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    classifier: InAlgorithm = SVM()
    assert classifier is not None
    assert classifier.name == "SVM"

    predictions: pd.DataFrame = classifier.run_test(new_train, new_test)
    assert predictions.values[predictions.values == 1].shape[0] == 193
    assert predictions.values[predictions.values == -1].shape[0] == 207


def test_threaded_custom_beutel():
    train, test = get_train_test()

    beut_model: PreAlgorithm = Beutel(epochs=5, fairness=FairType.EOPP)
    assert beut_model is not None
    assert beut_model.name == "Beutel"

    new_train_test_non_thread: Tuple[DataTuple, TestTuple] = beut_model.run(train, test)
    new_train_nt, new_test_nt = new_train_test_non_thread

    assert new_train_nt.x.shape[0] == train.x.shape[0]
    assert new_test_nt.x.shape[0] == test.x.shape[0]

    svm_model: InAlgorithm = SVM()
    assert svm_model is not None
    assert svm_model.name == "SVM"

    predictions = svm_model.run_test(new_train_nt, new_test_nt)
    assert predictions.values[predictions.values == 1].shape[0] == 202
    assert predictions.values[predictions.values == -1].shape[0] == 198

    model: PreAlgorithmAsync = Beutel(epochs=5, fairness=FairType.EOPP)
    assert model is not None
    assert model.name == "Beutel"

    new_train_test: Tuple[DataTuple, TestTuple] = run_blocking(
        model.run_async(train, test)
    )
    new_train, new_test = new_train_test

    assert new_train.x.shape[0] == train.x.shape[0]
    assert new_test.x.shape[0] == test.x.shape[0]

    classifier: InAlgorithm = SVM()
    assert classifier is not None
    assert classifier.name == "SVM"

    treaded_predictions: pd.DataFrame = classifier.run_test(new_train, new_test)
    assert treaded_predictions.values[treaded_predictions.values == 1].shape[0] == 202
    assert treaded_predictions.values[treaded_predictions.values == -1].shape[0] == 198



"""
Implementation of VFAE
"""
from typing import Any, Dict, List, Tuple

import pandas as pd
import torch
from torch import optim
from torch.utils.data import DataLoader

from ethicml.data import Adult, Compas, Credit, German, NonBinaryToy, Sqf, Toy
from ethicml.data.dataset import Dataset
from ethicml.implementations.pytorch_common import CustomDataset, TestDataset
from ethicml.implementations.utils import (
    pre_algo_argparser,
    load_data_from_flags,
    save_transformations,
)
from ethicml.implementations.vfae_modules.utils import loss_function
from ethicml.implementations.vfae_modules.vfae_network import VFAENetwork
from ethicml.utility.data_structures import DataTuple, TestTuple


def get_dataset_obj_by_name(name: str) -> Dataset:
    """
    Given a dataset name, get the corresponding dataset object
    Args:
        name:

    Returns:

    """
    lookup: Dict[str, Dataset] = {
        "Adult": Adult(),
        "Compas": Compas(),
        "Credit": Credit(),
        "German": German(),
        "NonBinaryToy": NonBinaryToy(),
        "SQF": Sqf(),
        "Toy": Toy(),
    }

    if name not in lookup:
        raise NotImplementedError("That dataset doesn't exist")

    return lookup[name]


def train_and_transform(
    train: DataTuple, test: TestTuple, flags: Any
) -> Tuple[DataTuple, TestTuple]:
    """
    train the model and transform the dataset
    Args:
        train:
        test:
        flags:

    Returns:
    """
    dataset = get_dataset_obj_by_name(flags['dataset'])

    # Set up the data
    train_data = CustomDataset(train)
    train_loader = DataLoader(train_data, batch_size=flags['batch_size'])

    test_data = TestDataset(test)
    test_loader = DataLoader(test_data, batch_size=flags['batch_size'])

    # Build Network
    model = VFAENetwork(
        dataset,
        flags['supervised'],
        train_data.size,
        latent_dims=50,
        z1_enc_size=flags['z1_enc_size'],
        z2_enc_size=flags['z2_enc_size'],
        z1_dec_size=flags['z1_dec_size'],
    ).to("cpu")
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    # Run Network
    for epoch in range(int(flags['epochs'])):
        train_model(epoch, model, train_loader, optimizer, flags)

    # Transform output
    post_train: List[List[float]] = []
    post_test: List[List[float]] = []
    model.eval()
    with torch.no_grad():
        for _x, _s, _ in train_loader:
            z1_mu, z1_logvar = model.encode_z1(_x, _s)
            z1 = model.reparameterize(z1_mu, z1_logvar)
            post_train += z1.data.tolist()
        for _x, _s in test_loader:
            z1_mu, z1_logvar = model.encode_z1(_x, _s)
            z1 = model.reparameterize(z1_mu, z1_logvar)
            post_test += z1.data.tolist()

    return (
        DataTuple(x=pd.DataFrame(post_train), s=train.s, y=train.y, name=f"VFAE: {train.name}"),
        TestTuple(x=pd.DataFrame(post_test), s=test.s, name=f"VFAE: {test.name}"),
    )


def train_model(epoch, model, train_loader, optimizer, flags):
    """
    Train the model
    Args:
        epoch:
        model:
        train_loader:
        optimizer:
        flags:

    Returns:

    """

    model.train()
    train_loss = 0
    for batch_idx, (data_x, data_s, data_y) in enumerate(train_loader):
        data_x = data_x.to("cpu")
        data_s = data_s.to("cpu")
        data_y = data_y.to("cpu")
        optimizer.zero_grad()
        z1_trip, z2_trip, z1_d_trip, x_dec, y_pred = model(data_x, data_s, data_y)

        data_trip = (data_x, data_s, data_y)

        loss_tuple = loss_function(flags, z1_trip, z2_trip, z1_d_trip, data_trip, x_dec, y_pred)
        prediction_loss, reconsruction_loss, kld_loss, mmd_loss = loss_tuple
        loss = kld_loss + reconsruction_loss + prediction_loss + mmd_loss
        loss.backward()
        # KLD_loss = KLD1 + KLD2
        train_loss += loss.item()
        optimizer.step()
        if batch_idx % 100 == 0:
            if flags['supervised']:
                print(
                    f'train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}'
                    f'({100. * batch_idx / len(train_loader):.0f}%)]\t'
                    f'Loss: {loss.item() / len(data_x):.6f}\t'
                    f'pred_loss: {prediction_loss.item():.6f}\t'
                    f'recon_loss: {reconsruction_loss.item():.6f}\t'
                    f'kld_loss: {kld_loss.item():.6f}\t'
                    f'mmd_loss: {flags["batch_size"] * mmd_loss.item():.6f}'
                )
            else:
                print(
                    f'train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}'
                    f'({100. * batch_idx / len(train_loader):.0f}%)]\t'
                    f'Loss: {loss.item() / len(data_x):.6f}\t'
                    f'recon_loss: {reconsruction_loss.item():.6f}\t'
                    f'mmd_loss: {flags["batch_size"] * mmd_loss.item():.6f}'
                )

    print(f'====> Epoch: {epoch} Average loss: {train_loss / len(train_loader.dataset):.4f}')


def main():
    """main method to run model"""
    parser = pre_algo_argparser()

    parser.add_argument("--supervised", type=bool, required=True)
    parser.add_argument("--fairness", type=str, required=True)
    parser.add_argument("--batch_size", type=int, required=True)
    parser.add_argument("--epochs", type=int, required=True)
    parser.add_argument("--dataset", type=str, required=True)
    parser.add_argument("--z1_enc_size", type=int, nargs="+", required=True)
    parser.add_argument("--z2_enc_size", type=int, nargs="+", required=True)
    parser.add_argument("--z1_dec_size", type=int, nargs="+", required=True)
    args = parser.parse_args()

    flags = vars(parser.parse_args())

    train, test = load_data_from_flags(flags)
    save_transformations(train_and_transform(train, test, flags), args)


if __name__ == "__main__":
    main()



"""Implementation of logistic regression (actually just a wrapper around sklearn)"""
from sklearn.linear_model import LogisticRegression

import pandas as pd

from ethicml.implementations.utils import InAlgoInterface


def train_and_predict(train, test, C):
    """Train a logistic regression model and compute predictions on the given test data"""
    clf = LogisticRegression(solver="liblinear", random_state=888, C=C, multi_class='auto')
    clf.fit(train.x, train.y.values.ravel())
    return pd.DataFrame(clf.predict(test.x), columns=["preds"])


def main():
    """This function runs the logistic regression model as a standalone program"""
    interface = InAlgoInterface()
    train, test = interface.load_data()
    C, = interface.remaining_args()
    interface.save_predictions(train_and_predict(train, test, C=float(C)))


if __name__ == "__main__":
    main()



"""
Returns a subset of the data. Used primarily in testing so that kernel methods finish in a
reasonable time
"""
from pathlib import Path
from typing import Tuple, List
from dataclasses import dataclass
from enum import Enum

import pandas as pd


@dataclass(frozen=True)  # "frozen" means the objects are immutable
class TestTuple:
    """A tuple of dataframes for the features and the sensitive attribute"""

    name: str  # name of the dataset
    x: pd.DataFrame  # features
    s: pd.DataFrame  # senstitive attributes


@dataclass(frozen=True)  # "frozen" means the objects are immutable
class DataTuple(TestTuple):
    """A tuple of dataframes for the features, the sensitive attribute and the class labels"""

    y: pd.DataFrame  # class labels


@dataclass(frozen=True)  # "frozen" means the objects are immutable
class TestPathTuple:
    """For algorithms that run in their own process, we pass around paths to the data"""

    x: Path  # path to file with features
    s: Path  # path to file with sensitive attributes
    name: Path  # name of the dataset


@dataclass(frozen=True)  # "frozen" means the objects are immutable
class PathTuple(TestPathTuple):
    """For algorithms that run in their own process, we pass around paths to the data"""

    y: Path  # path to file with class labels


def write_as_feather(
    train: DataTuple, test: TestTuple, data_dir: Path
) -> (Tuple[PathTuple, TestPathTuple]):
    """Write the given DataTuple to Feather files and return the file paths as PathTuples

    Args:
        train: tuple with training data
        test: tuple with test data
        data_dir: directory where the files should be stored
    Returns:
        tuple of tuple of paths (one tuple for training, one for test)
    """
    # create the directory if it doesn't already exist
    data_dir.mkdir(parents=True, exist_ok=True)

    def _save(data: pd.DataFrame, prefix: str, key: str) -> Path:
        # SUGGESTION: maybe the file names should be completely random to avoid collisions
        data_path = data_dir / f"data_{prefix}_{key}.feather"
        # write the file
        data.to_feather(data_path)
        return data_path

    train_paths = PathTuple(
        x=_save(train.x, "train", "x"),
        s=_save(train.s, "train", "s"),
        y=_save(train.y, "train", "y"),
        name=_save(pd.DataFrame([train.name], columns=['0']), "train", "name"),
    )
    test_paths = TestPathTuple(
        x=_save(test.x, "test", "x"),
        s=_save(test.s, "test", "s"),
        name=_save(pd.DataFrame([test.name], columns=['0']), "test", "name"),
    )
    return train_paths, test_paths


def apply_to_joined_tuple(mapper, datatup: DataTuple) -> DataTuple:
    """Concatenate the dataframes in a DataTuple and apply a function to it"""
    cols_x = datatup.x.columns
    cols_s = datatup.s.columns
    cols_y = datatup.y.columns
    joined = pd.concat([datatup.x, datatup.s, datatup.y], axis="columns", sort=False)
    joined = mapper(joined)
    return DataTuple(x=joined[cols_x], s=joined[cols_s], y=joined[cols_y], name=datatup.name)


def concat_dt(datatup_list: List[DataTuple], axis: str = "index", ignore_index: bool = False):
    """Concatenate the data tuples in the given list"""

    to_return = DataTuple(
        x=datatup_list[0].x, s=datatup_list[0].s, y=datatup_list[0].y, name=datatup_list[0].name
    )
    for i in range(1, len(datatup_list)):
        to_return = DataTuple(
            x=pd.concat(
                [to_return.x, datatup_list[i].x], axis=axis, sort=False, ignore_index=ignore_index
            ),
            s=pd.concat(
                [to_return.s, datatup_list[i].s], axis=axis, sort=False, ignore_index=ignore_index
            ),
            y=pd.concat(
                [to_return.y, datatup_list[i].y], axis=axis, sort=False, ignore_index=ignore_index
            ),
            name=to_return.name,
        )
    return to_return


def load_feather(output_path: Path) -> pd.DataFrame:
    """Load a dataframe from a feather file"""
    with output_path.open("rb") as file_obj:
        df = pd.read_feather(file_obj)
    return df


def get_subset(train: DataTuple, num: int = 500) -> DataTuple:
    """Get the first elements of the given dataset

    Args:
        train: training data

    Returns:
        subset of training data
    """
    return DataTuple(
        x=train.x.iloc[:num], s=train.s.iloc[:num], y=train.y.iloc[:num], name=train.name
    )


class FairType(Enum):
    """
    This is an enum that can be used to specify the type of fairness

    It basically works like the enums in C, but we can also give it values that are not integers.
    """

    DI = "DI"
    DP = "DI"  # alias for DI (for "demographic parity")
    EOPP = "Eq. Opp"
    EODDS = "Eq. Odds"

    def __str__(self) -> str:
        """This function is needed so that, for example, str(FairType.DI) returns 'DI'."""
        return self.value



"""
Simply returns the majority label from the train set
"""

import pandas as pd

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm


class Majority(InAlgorithm):
    """Simply returns the majority label from the train set"""

    def run(self, train, test):
        maj = train.y.mode().values
        return pd.DataFrame(maj.repeat(len(test.x)), columns=["preds"])

    @property
    def name(self) -> str:
        return "Majority"



"""
Logistic regaression with soft output
"""
from sklearn.linear_model import LogisticRegression

import pandas as pd

from ethicml.implementations.utils import InAlgoInterface


def train_and_predict(train, test, C):
    """Train a logistic regression model and compute predictions on the given test data"""
    clf = LogisticRegression(solver="liblinear", random_state=888, C=C, multi_class='auto')

    clf.fit(train.x, train.y.values.ravel())
    return pd.DataFrame(clf.predict_proba(test.x)[:, 1], columns=["preds"])


def main():
    """main method to run model"""
    interface = InAlgoInterface()
    train, test = interface.load_data()
    C, = interface.remaining_args()
    interface.save_predictions(train_and_predict(train, test, float(C)))


if __name__ == "__main__":
    main()



"""
Set of scripts for splitting the train and test datasets based on conditions
"""

import re
from typing import Tuple
from functools import partial
import pandas as pd
from ethicml.utility.data_structures import DataTuple, apply_to_joined_tuple


def dataset_from_cond(dataset: pd.DataFrame, cond: str):
    """Return the dataframe that meets some condition"""
    original_column_names = dataset.columns
    # make column names query-friendly
    dataset = dataset.rename(axis="columns", mapper=_make_valid_variable_name)
    subset = dataset.query(cond)
    subset.columns = original_column_names
    return subset


def query_dt(datatup: DataTuple, query_str: str):
    """Query a datatuple"""
    assert isinstance(query_str, str)
    assert isinstance(datatup, DataTuple)
    query_func = partial(dataset_from_cond, cond=query_str)
    return apply_to_joined_tuple(query_func, datatup)


def domain_split(datatup: DataTuple, tr_cond: str, te_cond: str) -> Tuple[DataTuple, DataTuple]:
    """
    Splits a datatuple based on a condition
    Args:
        datatup: DataTuple
        tr_cond: condition for the training set
        te_cond: condition for the test set

    Returns: Tuple of DataTuple split into train and test. The test is all those that meet
    the test condition plus the same percentage again of the train set.
    """
    dataset = datatup.x

    train_dataset = dataset_from_cond(dataset, tr_cond)
    test_dataset = dataset_from_cond(dataset, te_cond)

    assert train_dataset.shape[0] + test_dataset.shape[0] == dataset.shape[0]

    test_pct = test_dataset.shape[0] / dataset.shape[0]
    train_pct = 1 - test_pct

    train_train_pcnt = (1 - (test_pct * 2)) / train_pct

    train_train = train_dataset.sample(frac=train_train_pcnt, random_state=888)
    test_train = train_dataset.drop(train_train.index)

    test = pd.concat([test_train, test_dataset], axis="index")

    train_x = datatup.x.loc[train_train.index].reset_index(drop=True)
    train_s = datatup.s.loc[train_train.index].reset_index(drop=True)
    train_y = datatup.y.loc[train_train.index].reset_index(drop=True)

    train_datatup = DataTuple(x=train_x, s=train_s, y=train_y, name=datatup.name)

    test_x = datatup.x.loc[test.index].reset_index(drop=True)
    test_s = datatup.s.loc[test.index].reset_index(drop=True)
    test_y = datatup.y.loc[test.index].reset_index(drop=True)

    test_datatup = DataTuple(x=test_x, s=test_s, y=test_y, name=datatup.name)

    return train_datatup, test_datatup


def _make_valid_variable_name(name: str) -> str:
    """Convert a string into a valid Python variable name"""
    # Ensure that it's a string
    name = str(name)

    # Python variable names may only contain digits, letters and underscores
    explicit_substitutions = {
        "-": "_",  # Replace hyphens with underscores
        "+": "_plus_",
        "=": "_eq_",
    }
    for original, replacement in explicit_substitutions.items():
        name = name.replace(original, replacement)

    # Remove every other disallowed character
    name = re.sub("[^0-9a-zA-Z_]", "", name)

    # Python variables must start with a letter or an underscore
    # Thus, we add an underscore, if the string starts with a digit
    if name[0] in ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9"):
        name = "_" + name
    return name



""""
Cross Validation for any in process (at the moment) Algoruthm"""


from collections import defaultdict
from itertools import product
from operator import itemgetter
from statistics import mean
from typing import Dict, List, Tuple, Any
from typing import Type

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.utility.data_structures import DataTuple
from ethicml.metrics import Accuracy, Metric, CV
from ethicml.preprocessing.train_test_split import fold_data


class Results:
    """
    Stores the results of the experiments
    """

    def __init__(self):
        self.storage: List[Tuple[Dict[str, Any], int, str, float]] = []

    def append(self, config: Dict[str, Any], fold_id: int, score_name: str, score: float) -> None:
        self.storage.append((config, fold_id, score_name, score))

    def get_params_for_best(self, measure: Metric) -> Dict[str, Any]:
        """
        Get the hyperparameter combination for the best performance of a measure
        Args:
            measure:

        Returns:

        """
        candidates = [val for val in self.storage if val[2] == measure.name]

        grouped: Dict[str, List[Any]] = defaultdict(list)
        for tup in candidates:
            hyp_dic: Dict[str, Any] = tup[0]
            str_hyp_dict = ", ".join("{!s}={!r}".format(key, val) for (key, val) in hyp_dic.items())
            grouped[str_hyp_dict].append(tup[1:])

        mean_vals: Dict[str, float] = defaultdict(float)
        for exp, results in grouped.items():
            results = [list(res) for res in results]
            mean_vals[exp] = mean([res[2] for res in results])

        best_params = (max(mean_vals.items(), key=itemgetter(1)))[0]

        params_for_best = candidates[0][0]
        for k in candidates:
            hyp_dic = k[0]
            str_hyp_dict = ", ".join("{!s}={!r}".format(key, val) for (key, val) in hyp_dic.items())
            if str_hyp_dict == best_params:
                params_for_best = hyp_dic
                break
        return params_for_best


class CrossValidator:
    """
    Object used to run cross-validatioon on a model
    """

    def __init__(self, model: Type[InAlgorithm], hyperparams: Dict[str, List[Any]], folds: int = 3):
        self.model = model
        self.hyperparams = hyperparams
        self.results = Results()
        self.folds = folds

        keys, values = zip(*hyperparams.items())
        self.experiments = [dict(zip(keys, v)) for v in product(*values)]

    def run(self, train: DataTuple, measures=None):
        """
        runs the cross validation experiments
        Args:
            train:
            measures:

        Returns:

        """
        if measures is None:
            measures = [Accuracy(), CV()]

        for i, (train_fold, val) in enumerate(fold_data(train, folds=self.folds)):
            for experiment in self.experiments:
                model = self.model(**experiment)
                preds = model.run(train_fold, val)
                for measure in measures:
                    self.results.append(experiment, i, measure.name, measure.score(preds, val))
                print(f"fold_{i}_model_{model.name}_completed")

    def best(self, measure) -> InAlgorithm:
        return self.model(**self.results.get_params_for_best(measure))

    def best_hyper_params(self, measure):
        return self.results.get_params_for_best(measure)



"""
creates plots of a dataset
"""


import os
from pathlib import Path
from typing import Tuple, List

import imageio
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib as mpl

from ethicml.utility.data_structures import DataTuple

MARKERS = ["s", "p", "P", "*", "+", "x", "o", "v"]


def save_2d_plot(data: DataTuple, filepath: str):
    """

    Args:
        data:
        filepath:
    """
    file_path = Path(filepath)
    columns = data.x.columns

    amalgamated = pd.concat([data.x, data.s, data.y], axis="columns")

    plot = sns.scatterplot(
        x=columns[0],
        y=columns[1],
        hue=data.y.columns[0],
        palette="Set2",
        data=amalgamated,
        style=data.s.columns[0],
        legend="full",
    )

    if not os.path.exists(file_path.parent):
        os.mkdir(file_path.parent)
    plot.figure.savefig(file_path)
    plt.clf()


def save_jointplot(data: DataTuple, filepath: str, dims: Tuple[int, int] = (0, 1)):
    """

    Args:
        data:
        filepath:
        dims:
    """
    file_path = Path(filepath)
    columns = data.x.columns

    amalgamated = pd.concat([data.x, data.y], axis="columns")

    plot = sns.jointplot(x=columns[dims[0]], y=columns[dims[1]], data=amalgamated, kind="kde")

    if not os.path.exists(file_path.parent):
        os.mkdir(file_path)
    plot.savefig(file_path)
    plt.clf()


def make_gif(files: List[str], name="movie"):
    """

    Args:
        files:
        name:
    """
    images = []
    for filename in files:
        images.append(imageio.imread(filename))
    imageio.mimsave(f"{name}.gif", images)


def save_label_plot(data: DataTuple, filename: str) -> None:
    """

    Args:
        data:
        filename:
    """
    file_path = Path(filename)

    # Only consider 1 sens attr for now
    s_col = data.s.columns[0]
    s_values = data.s[s_col].value_counts() / data.s[s_col].count()

    s_0_val = s_values[0]
    s_1_val = s_values[1]

    s_0_label = s_values.index[0]
    s_1_label = s_values.index[1]

    y_col = data.y.columns[0]
    y_s0 = (
        data.y[y_col][data.s[s_col] == 0].value_counts() / data.y[y_col][data.s[s_col] == 0].count()
    )
    y_s1 = (
        data.y[y_col][data.s[s_col] == 1].value_counts() / data.y[y_col][data.s[s_col] == 1].count()
    )

    y_0_label = y_s0.index[0]
    y_1_label = y_s0.index[1]

    mpl.style.use("seaborn-pastel")
    # plt.xkcd()

    # fig, ax = plt.subplots()

    _ = plt.bar(
        0,
        height=y_s0[y_0_label] * 100,
        width=s_0_val * 100,
        align="edge",
        edgecolor="black",
        color="C0",
    )
    _ = plt.bar(
        s_0_val * 100,
        height=y_s1[y_0_label] * 100,
        width=s_1_val * 100,
        align="edge",
        edgecolor="black",
        color="C1",
    )
    _ = plt.bar(
        0,
        height=y_s0[y_1_label] * 100,
        width=s_0_val * 100,
        bottom=y_s0[y_0_label] * 100,
        align="edge",
        edgecolor="black",
        color="C2",
    )
    _ = plt.bar(
        s_0_val * 100,
        height=y_s1[y_1_label] * 100,
        width=s_1_val * 100,
        bottom=y_s1[y_0_label] * 100,
        align="edge",
        edgecolor="black",
        color="C3",
    )

    plt.ylim([0, 100])
    plt.xlim([0, 100])
    plt.ylabel(f"Percent {y_col}=y")
    plt.xlabel(f"Percent {s_col}=s")
    plt.title("Dataset Composition by class and sensitive attribute")

    plt.legend(
        [
            f"y={y_0_label}, s={s_0_label}",
            f"y={y_0_label}, s={s_1_label}",
            f"y={y_1_label}, s={s_0_label}",
            f"y={y_1_label}, s={s_1_label}",
        ]
    )

    if not os.path.exists(file_path.parent):
        os.mkdir(file_path)
    plt.savefig(file_path)
    plt.clf()



"""
Base class for Algorithms
"""
import sys
from pathlib import Path
from abc import ABC, abstractmethod, ABCMeta
from typing import List, Optional, Dict, Coroutine, TypeVar, Any
import asyncio

import pandas as pd


def load_dataframe(path: Path) -> pd.DataFrame:
    """Load dataframe from a feather file"""
    with path.open("rb") as file:
        df = pd.read_feather(file)
    return df


class Algorithm(ABC):
    """Base class for Algorithms"""

    def __init__(self, *args, **kwargs):  # this seems really pointless but it makes mypy happy
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the algorithm"""


class AlgorithmAsync(Algorithm, metaclass=ABCMeta):
    """Base class of async methods. This class is meant to be used in conjuction with `Algorithm`"""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the algorithm"""

    @property
    def _executable(self) -> str:
        """
        Path to a (Python) executable

        By default, the Python executable that called this script is used.
        """
        return sys.executable

    async def _call_script(self, cmd_args: List[str], env: Optional[Dict[str, str]] = None):
        """This function calls a (Python) script as a separate process

        An exception is thrown if the called script failed.

        Args:
            cmd_args: list of strings that are passed as commandline arguments to the executable
            env: environment variables specified as a dictionary; e.g. {"PATH": "/usr/bin"}
        """
        process = await asyncio.create_subprocess_exec(  # wait for process creation to finish
            self._executable,
            *cmd_args,
            stderr=asyncio.subprocess.PIPE,  # we capture the stderr for errors
            env=env,
        )
        # print(f"Started: {cmd_args!r} (pid = {process.pid})")

        try:  # wait for process itself to finish
            one_hour = 3600
            _, stderr = await asyncio.wait_for(process.communicate(), one_hour)
        except asyncio.TimeoutError as error:
            raise RuntimeError("The script timed out.") from error

        if process.returncode == 0:
            print(f"Success: {cmd_args!r} (pid = {process.pid})")
        else:
            print(f"Failure: {cmd_args!r} (pid = {process.pid})")
            if stderr:
                print(stderr.decode().strip())
            raise RuntimeError(
                f"The script failed. Supplied arguments: {cmd_args} with exec: {self._executable}"
            )


_T = TypeVar("_T")


def run_blocking(promise: Coroutine[Any, Any, _T]) -> _T:
    """Run an asynchronous process as a blocking process"""
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(promise)



"""
Test the saving data capability
"""
import pandas as pd
import numpy as np

from ethicml.utility.data_structures import DataTuple
from ethicml.algorithms.algorithm_base import run_blocking
from ethicml.algorithms.inprocess import InAlgorithmAsync


def test_simple_saving():
    data_tuple = DataTuple(
        x=pd.DataFrame(
            {'a1': np.array([3.2, 9.4, np.nan, 0.0]), 'a2': np.array([1, 1, 0, 1])}),
        s=pd.DataFrame(
            {'b1': np.array([1.8, -0.3, 1e10]),
             'b2': np.array([1, 1, -1]),
             'b3': np.array([0, 1, 0])}),
        y=pd.DataFrame(
            {'c1': np.array([-2, -3, np.nan]),
             'c3': np.array([0, 1, 0])}),
        name='test_data'
    )

    class CheckEquality(InAlgorithmAsync):
        """Dummy algorithm class for testing whether writing and reading feather files works"""

        def name(self):
            return "Check equality"

        def _script_command(self, train_paths, _, pred_path):
            """Check if the dataframes loaded from the files are the same as the original ones"""
            x_loaded = pd.read_feather(train_paths.x)
            s_loaded = pd.read_feather(train_paths.s)
            y_loaded = pd.read_feather(train_paths.y)
            pd.testing.assert_frame_equal(data_tuple.x, x_loaded)
            pd.testing.assert_frame_equal(data_tuple.s, s_loaded)
            pd.testing.assert_frame_equal(data_tuple.y, y_loaded)
            # the following command copies the x of the training data to the pred_path location
            return ['-c', f'import shutil; shutil.copy("{train_paths.x}", "{pred_path}")']

    data_x = run_blocking(CheckEquality().run_async(data_tuple, data_tuple))
    pd.testing.assert_frame_equal(data_tuple.x, data_x)



"""
Class to describe features of the Test dataset
"""
from typing import List, Dict
from .dataset import Dataset


class Toy(Dataset):
    """Class to describe features of the Test dataset"""

    cont_features: List[str]
    disc_features: List[str]

    def __init__(self):
        super().__init__()
        self.cont_features = ["a1", "a2"]
        self.disc_features = []

    @property
    def name(self) -> str:
        return "Toy"

    @property
    def filename(self) -> str:
        return "toy.csv"

    @property
    def feature_split(self) -> Dict[str, List[str]]:
        return {"x": ["a1", "a2"], "s": ["s"], "y": ["y"]}



"""
Test preprocessing capabilities
"""
import math

from typing import Tuple

from ethicml.utility.data_structures import DataTuple
from ethicml.data.load import load_data
from ethicml.data.toy import Toy
from ethicml.preprocessing.train_test_split import train_test_split


def test_train_test_split():
    data: DataTuple = load_data(Toy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    assert train is not None
    assert test is not None
    assert train.x.shape[0] > test.x.shape[0]
    assert train.x['a1'].values[0] == 2.7839204628851526
    assert train.x['a2'].values[0] == 6.083679468491366

    assert train.x.shape[0] == train.s.shape[0]
    assert train.s.shape[0] == train.y.shape[0]

    len_default = math.floor((2000 / 100) * 80)
    assert train.s.shape[0] == len_default
    assert test.s.shape[0] == 2000 - len_default

    len_0_9 = math.floor((2000 / 100) * 90)
    train, test = train_test_split(data, train_percentage=0.9)
    assert train.s.shape[0] == len_0_9
    assert test.s.shape[0] == 2000 - len_0_9

    len_0_7 = math.floor((2000 / 100) * 70)
    train, test = train_test_split(data, train_percentage=0.7)
    assert train.s.shape[0] == len_0_7
    assert test.s.shape[0] == 2000 - len_0_7

    len_0_5 = math.floor((2000 / 100) * 50)
    train, test = train_test_split(data, train_percentage=0.5)
    assert train.s.shape[0] == len_0_5
    assert test.s.shape[0] == 2000 - len_0_5

    len_0_3 = math.floor((2000 / 100) * 30)
    train, test = train_test_split(data, train_percentage=0.3)
    assert train.s.shape[0] == len_0_3
    assert test.s.shape[0] == 2000 - len_0_3

    len_0_1 = math.floor((2000 / 100) * 10)
    train, test = train_test_split(data, train_percentage=0.1)
    assert train.s.shape[0] == len_0_1
    assert test.s.shape[0] == 2000 - len_0_1

    len_0_0 = math.floor((2000 / 100) * 0)
    train, test = train_test_split(data, train_percentage=0.0)
    assert train.s.shape[0] == len_0_0
    assert "Toy - Train" == train.name
    assert test.s.shape[0] == 2000 - len_0_0
    assert "Toy - Test" == test.name


def test_random_seed():
    data: DataTuple = load_data(Toy())
    train_test_0: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train_0, test_0 = train_test_0
    assert train_0 is not None
    assert test_0 is not None
    assert train_0.x.shape[0] > test_0.x.shape[0]
    assert train_0.x['a1'].values[0] == 2.7839204628851526
    assert train_0.x['a2'].values[0] == 6.083679468491366

    assert train_0.x.shape[0] == train_0.s.shape[0]
    assert train_0.s.shape[0] == train_0.y.shape[0]

    train_test_1: Tuple[DataTuple, DataTuple] = train_test_split(data, random_seed=1)
    train_1, test_1 = train_test_1
    assert train_1 is not None
    assert test_1 is not None
    assert train_1.x.shape[0] > test_1.x.shape[0]
    assert train_1.x['a1'].values[0] == 2.6158100167119773
    assert train_1.x['a2'].values[0] == -2.026281606912103

    assert train_1.x.shape[0] == train_1.s.shape[0]
    assert train_1.s.shape[0] == train_1.y.shape[0]

    train_test_2: Tuple[DataTuple, DataTuple] = train_test_split(data, random_seed=2)
    train_2, test_2 = train_test_2
    assert train_2 is not None
    assert test_2 is not None
    assert train_2.x.shape[0] > test_2.x.shape[0]
    assert train_2.x['a1'].values[0] == 0.6431486026206228
    assert train_2.x['a2'].values[0] == -0.09879806963941018

    assert train_2.x.shape[0] == train_2.s.shape[0]
    assert train_2.s.shape[0] == train_2.y.shape[0]

    train_test_3: Tuple[DataTuple, DataTuple] = train_test_split(data, random_seed=3)
    train_3, test_3 = train_test_3
    assert train_3 is not None
    assert test_3 is not None
    assert train_3.x.shape[0] > test_3.x.shape[0]
    assert train_3.x['a1'].values[0] == 0.8165458710908045
    assert train_3.x['a2'].values[0] == -5.456548268244318

    assert train_3.x.shape[0] == train_3.s.shape[0]
    assert train_3.s.shape[0] == train_3.y.shape[0]



from typing import List, Dict, Any
import pandas as pd
import pytest
from pytest import approx

from ethicml.algorithms.algorithm_base import run_blocking
from ethicml.algorithms.inprocess import (
    Agarwal,
    GPyT,
    GPyTDemPar,
    GPyTEqOdds,
    InAlgorithm,
    InAlgorithmAsync,
    Kamishima,
    LRCV,
    LRProb,
    LR,
    SVM,
    Kamiran,
    Majority,
    MLP,
)
from ethicml.evaluators.cross_validator import CrossValidator
from ethicml.metrics import Accuracy
from ethicml.utility.heaviside import Heaviside
from tests.run_algorithm_test import get_train_test


def test_svm():
    train, test = get_train_test()

    model: InAlgorithm = SVM()
    assert model is not None
    assert model.name == "SVM"

    predictions: pd.DataFrame = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 201
    assert predictions.values[predictions.values == -1].shape[0] == 199


def test_majority():
    train, test = get_train_test()

    model: InAlgorithm = Majority()
    assert model is not None
    assert model.name == "Majority"

    predictions: pd.DataFrame = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 0
    assert predictions.values[predictions.values == -1].shape[0] == 400


def test_mlp():
    train, test = get_train_test()

    model: InAlgorithm = MLP()
    assert model is not None
    assert model.name == "MLP"

    predictions: pd.DataFrame = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 200
    assert predictions.values[predictions.values == -1].shape[0] == 200


def test_cv_svm():
    train, test = get_train_test()

    hyperparams: Dict[str, List[Any]] = {'C': [1, 10, 100], 'kernel': ['rbf', 'linear']}

    svm_cv = CrossValidator(SVM, hyperparams, folds=3)

    assert svm_cv is not None
    assert isinstance(svm_cv.model(), InAlgorithm)

    svm_cv.run(train)

    best_model = svm_cv.best(Accuracy())

    predictions: pd.DataFrame = best_model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189


def test_cv_lr():
    train, test = get_train_test()

    hyperparams: Dict[str, List[int]] = {'C': [1, 10, 100]}

    lr_cv = CrossValidator(LR, hyperparams, folds=3)

    assert lr_cv is not None
    assert isinstance(lr_cv.model(), InAlgorithm)

    lr_cv.run(train)

    best_model = lr_cv.best(Accuracy())

    predictions: pd.DataFrame = best_model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189


@pytest.fixture(scope="module")
def kamishima():
    kamishima_algo = Kamishima()
    yield kamishima_algo
    print("teardown Kamishima")
    kamishima_algo.remove()


def test_kamishima(kamishima):
    train, test = get_train_test()

    model: InAlgorithmAsync = kamishima
    assert model.name == "Kamishima"

    assert model is not None

    predictions: pd.DataFrame = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 208
    assert predictions.values[predictions.values == -1].shape[0] == 192


# @pytest.fixture(scope="module")
# def gpyt_models():
#     # the gpyt models are created together because they share a git repository
#     gpyt = GPyT(epochs=2, length_scale=2.4)
#     gpyt_dem_par = GPyTDemPar(epochs=2, length_scale=0.05)
#     gpyt_eq_odds = GPyTEqOdds(epochs=2, tpr=1.0, length_scale=0.05)
#     yield (gpyt, gpyt_dem_par, gpyt_eq_odds)
#     gpyt.remove()  # only has to be called from one of the models because they share a directory
#
#
# def test_gpyt(gpyt_models):
#     train, test = get_train_test()
#
#     baseline: InAlgorithmAsync = gpyt_models[0]
#     dem_par: InAlgorithmAsync = gpyt_models[1]
#     eq_odds: InAlgorithmAsync = gpyt_models[2]
#
#     assert baseline.name == "GPyT_in_True"
#     predictions: pd.DataFrame = run_blocking(baseline.run_async(train, test))
#     assert predictions.values[predictions.values == 1].shape[0] == approx(210, rel=0.1)
#     assert predictions.values[predictions.values == -1].shape[0] == approx(190, rel=0.1)
#
#     assert dem_par.name == "GPyT_dem_par_in_True"
#     predictions = run_blocking(dem_par.run_async(train, test))
#     assert predictions.values[predictions.values == 1].shape[0] == approx(182, rel=0.1)
#     assert predictions.values[predictions.values == -1].shape[0] == approx(218, rel=0.1)
#
#     assert eq_odds.name == "GPyT_eq_odds_in_True_tpr_1.0"
#     predictions = run_blocking(eq_odds.run_async(train, test))
#     assert predictions.values[predictions.values == 1].shape[0] == approx(179, rel=0.1)
#     assert predictions.values[predictions.values == -1].shape[0] == approx(221, rel=0.1)


def test_threaded_svm():
    train, test = get_train_test()

    model: InAlgorithmAsync = SVM()
    assert model is not None
    assert model.name == "SVM"

    predictions: pd.DataFrame = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 201
    assert predictions.values[predictions.values == -1].shape[0] == 199

    predictions_non_threaded: pd.DataFrame = model.run(train, test)
    assert predictions_non_threaded.values[predictions_non_threaded.values == 1].shape[0] == 201
    assert predictions_non_threaded.values[predictions_non_threaded.values == -1].shape[0] == 199


def test_lr():
    train, test = get_train_test()

    model: InAlgorithm = LR()
    assert model is not None
    assert model.name == "Logistic Regression"

    predictions: pd.DataFrame = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189


def test_agarwal():
    train, test = get_train_test()
    model: InAlgorithm = Agarwal()
    assert model is not None
    assert model.name == "Agarwal LR"

    predictions: pd.DataFrame = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 145
    assert predictions.values[predictions.values == -1].shape[0] == 255

    model = Agarwal(fairness="EqOd")
    predictions = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 162
    assert predictions.values[predictions.values == -1].shape[0] == 238

    model = Agarwal(classifier="SVM")
    predictions = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 141
    assert predictions.values[predictions.values == -1].shape[0] == 259

    model = Agarwal(classifier="SVM", kernel='linear')
    predictions = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189

    model = Agarwal(classifier="SVM", fairness="EqOd")
    predictions = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 159
    assert predictions.values[predictions.values == -1].shape[0] == 241

    model = Agarwal(classifier="SVM", fairness="EqOd", kernel="linear")
    predictions = model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189

    model = Agarwal(classifier="SVM", fairness="EqOd")
    predictions = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 159
    assert predictions.values[predictions.values == -1].shape[0] == 241


def test_threaded_lr():
    train, test = get_train_test()

    model: InAlgorithmAsync = LR()
    assert model.name == "Logistic Regression"

    predictions: pd.DataFrame = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189

    predictions_non_threaded: pd.DataFrame = model.run(train, test)
    assert predictions_non_threaded.values[predictions_non_threaded.values == 1].shape[0] == 211
    assert predictions_non_threaded.values[predictions_non_threaded.values == -1].shape[0] == 189


def test_threaded_lr_cross_validated():
    train, test = get_train_test()

    model: InAlgorithmAsync = LRCV()
    assert model.name == "LRCV"

    predictions: pd.DataFrame = run_blocking(model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189

    predictions_non_threaded: pd.DataFrame = model.run(train, test)
    assert predictions_non_threaded.values[predictions_non_threaded.values == 1].shape[0] == 211
    assert predictions_non_threaded.values[predictions_non_threaded.values == -1].shape[0] == 189


def test_threaded_lr_prob():
    train, test = get_train_test()

    model: InAlgorithmAsync = LRProb()
    assert model.name == "Logistic Regression Prob"

    heavi = Heaviside()

    predictions_non_threaded: pd.DataFrame = model.run(train, test)
    predictions_non_threaded = predictions_non_threaded.apply(heavi.apply)
    predictions_non_threaded = predictions_non_threaded.replace(0, -1)

    assert predictions_non_threaded.values[predictions_non_threaded.values == 1].shape[0] == 211
    assert predictions_non_threaded.values[predictions_non_threaded.values == -1].shape[0] == 189

    predictions: pd.DataFrame = run_blocking(model.run_async(train, test))
    predictions = predictions.apply(heavi.apply)
    predictions = predictions.replace(0, -1)
    assert predictions.values[predictions.values == 1].shape[0] == 211
    assert predictions.values[predictions.values == -1].shape[0] == 189


def test_kamiran():
    train, test = get_train_test()

    kamiran_model: InAlgorithmAsync = Kamiran()
    assert kamiran_model is not None
    assert kamiran_model.name == "Kamiran & Calders LR"

    predictions: pd.DataFrame = kamiran_model.run(train, test)
    assert predictions.values[predictions.values == 1].shape[0] == 210
    assert predictions.values[predictions.values == -1].shape[0] == 190

    predictions: pd.DataFrame = run_blocking(kamiran_model.run_async(train, test))
    assert predictions.values[predictions.values == 1].shape[0] == 210
    assert predictions.values[predictions.values == -1].shape[0] == 190



"""
Simplae class to make class labels binary. Useful if a network uses BCELoss for example
"""

import numpy as np

from ethicml.utility.data_structures import DataTuple


def assert_binary_labels(train: DataTuple):
    """
    Assert that datasets only include binary labels
    Args:
        train:

    Returns:

    """

    y_col = train.y.columns[0]
    assert train.y[y_col].nunique() == 2
    assert (np.unique(train.y[y_col].values) == np.array([0, 1])).all()


class LabelBinarizer:
    """
    If a dataset has labels [-1,1], then this will make it so the labels = [0,1]
    """

    def __init__(self):
        self.min_val: int
        self.max_val: int

    def adjust(self, dataset: DataTuple) -> DataTuple:
        """
        Take a datatuple and make the labels [0,1]
        Args:
            dataset:

        Returns:

        """
        y_col = dataset.y.columns[0]
        assert dataset.y[y_col].nunique() == 2

        # make copy of dataset
        dataset = DataTuple(x=dataset.x, s=dataset.s, y=dataset.y.copy(), name=dataset.name)

        self.min_val = dataset.y.values.min()
        self.max_val = dataset.y.values.max()

        y_col = dataset.y.columns[0]

        dataset.y[y_col].replace(self.min_val, 0, inplace=True)
        dataset.y[y_col].replace(self.max_val, 1, inplace=True)

        return DataTuple(x=dataset.x, s=dataset.s, y=dataset.y, name=dataset.name)

    def post(self, dataset: DataTuple) -> DataTuple:
        """
        Inverse of adjust
        Args:
            dataset:

        Returns:

        """
        y_col = dataset.y.columns[0]
        assert dataset.y[y_col].nunique() == 2

        # make copy of dataset
        dataset = DataTuple(x=dataset.x, s=dataset.s, y=dataset.y.copy(), name=dataset.name)

        y_col = dataset.y.columns[0]

        dataset.y[y_col].replace(0, self.min_val, inplace=True)
        dataset.y[y_col].replace(1, self.max_val, inplace=True)

        return DataTuple(x=dataset.x, s=dataset.s, y=dataset.y, name=dataset.name)



"""
For assessing Nomralized Mutual Information
"""

import pandas as pd

from sklearn.metrics import normalized_mutual_info_score as nmis

from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class NMI(Metric):
    """Normalized Mutual Information"""

    def __init__(self, pos_class: int = 1, base: str = "y"):
        super().__init__(pos_class=pos_class)
        if base not in ["s", "y"]:
            raise NotImplementedError("Can only calculate NMI of predictions with regard to y or s")
        self.base = base

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        if self.base == "y":
            base_values = actual.y.values.flatten()
        else:
            base_values = actual.s.values.flatten()
        return nmis(base_values, prediction.values.flatten(), average_method="geometric")

    @property
    def name(self) -> str:
        return f"NMI preds and {self.base}"



"""Beutel's algorithm"""
from pathlib import Path
from typing import List, Sequence, Tuple

from ethicml.utility.data_structures import DataTuple, TestTuple, PathTuple, TestPathTuple, FairType
from .pre_algorithm import PreAlgorithmAsync
from .interface import flag_interface


class Beutel(PreAlgorithmAsync):
    """Beutel's adversarially learned fair representations"""

    def __init__(
        self,
        fairness: FairType = FairType.DI,
        enc_size: Sequence[int] = (40,),
        adv_size: Sequence[int] = (40,),
        pred_size: Sequence[int] = (40,),
        enc_activation: str = "Sigmoid()",
        adv_activation: str = "Sigmoid()",
        batch_size: int = 64,
        y_loss: str = "BCELoss()",
        s_loss: str = "BCELoss()",
        epochs: int = 50,
        adv_weight: float = 1.0,
        validation_pcnt: float = 0.1,
    ):
        # pylint: disable=too-many-arguments
        super().__init__()
        self.fairness = fairness
        self.enc_size = enc_size
        self.adv_size = adv_size
        self.pred_size = pred_size
        self.enc_activation = enc_activation
        self.adv_activation = adv_activation
        self.batch_size = batch_size
        self.y_loss = y_loss
        self.s_loss = s_loss
        self.epochs = epochs
        self.adv_weight = adv_weight
        self.validation_pcnt = validation_pcnt

    def run(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        from ...implementations import beutel  # only import this on demand because of pytorch

        # SUGGESTION: it would be great if BeutelSettings could already be created in the init
        flags = beutel.BeutelSettings(
            fairness=self.fairness,
            enc_size=self.enc_size,
            adv_size=self.adv_size,
            pred_size=self.pred_size,
            enc_activation=self.enc_activation,
            adv_activation=self.adv_activation,
            batch_size=self.batch_size,
            y_loss=self.y_loss,
            s_loss=self.s_loss,
            epochs=self.epochs,
            adv_weight=self.adv_weight,
            validation_pcnt=self.validation_pcnt,
        )
        return beutel.train_and_transform(train, test, flags)

    def _script_command(
        self,
        train_paths: PathTuple,
        test_paths: TestPathTuple,
        new_train_x_path: Path,
        new_train_s_path: Path,
        new_train_y_path: Path,
        new_train_name_path: Path,
        new_test_x_path: Path,
        new_test_s_path: Path,
        new_test_name_path: Path,
    ) -> List[str]:
        args = flag_interface(
            train_paths,
            test_paths,
            new_train_x_path,
            new_train_s_path,
            new_train_y_path,
            new_train_name_path,
            new_test_x_path,
            new_test_s_path,
            new_test_name_path,
            vars(self),
        )
        return ["-m", "ethicml.implementations.beutel"] + args

    @property
    def name(self) -> str:
        return "Beutel"



"""
Implementatioon of Kamiran and Calders 2012

Heavily based on AIF360
https://github.com/IBM/AIF360/blob/master/aif360/algorithms/preprocessing/reweighing.py

    References:
        .. [4] F. Kamiran and T. Calders,  "Data Preprocessing Techniques for
           Classification without Discrimination," Knowledge and Information
           Systems, 2012.
"""

import pandas as pd
import numpy as np
from sklearn.linear_model import LogisticRegression

from ethicml.utility.data_structures import DataTuple
from ethicml.implementations.utils import InAlgoInterface
from ethicml.implementations.svm import select_svm


def _obtain_conditionings(dataset: DataTuple):
    """Obtain the necessary conditioning boolean vectors to compute instance level weights."""
    y_col = dataset.y.columns[0]
    y_pos = dataset.y[y_col].max()
    y_neg = dataset.y[y_col].min()
    s_col = dataset.s.columns[0]
    s_pos = dataset.s[s_col].max()
    s_neg = dataset.s[s_col].min()

    # combination of label and privileged/unpriv. groups
    cond_p_fav = dataset.x.loc[(dataset.y[y_col] == y_pos) & (dataset.s[s_col] == s_pos)]
    cond_p_unfav = dataset.x.loc[(dataset.y[y_col] == y_neg) & (dataset.s[s_col] == s_pos)]
    cond_up_fav = dataset.x.loc[(dataset.y[y_col] == y_pos) & (dataset.s[s_col] == s_neg)]
    cond_up_unfav = dataset.x.loc[(dataset.y[y_col] == y_neg) & (dataset.s[s_col] == s_neg)]

    return cond_p_fav, cond_p_unfav, cond_up_fav, cond_up_unfav


def compute_weights(train: DataTuple) -> pd.DataFrame:
    """Compute weights for all samples"""
    np.random.seed(888)
    (cond_p_fav, cond_p_unfav, cond_up_fav, cond_up_unfav) = _obtain_conditionings(train)

    y_col = train.y.columns[0]
    y_pos = train.y[y_col].max()
    y_neg = train.y[y_col].min()
    s_col = train.s.columns[0]
    s_pos = train.s[s_col].max()
    s_neg = train.s[s_col].min()

    num_samples = train.x.shape[0]
    n_p = train.s.loc[train.s[s_col] == s_pos].shape[0]
    n_up = train.s.loc[train.s[s_col] == s_neg].shape[0]
    n_fav = train.y.loc[train.y[y_col] == y_pos].shape[0]
    n_unfav = train.y.loc[train.y[y_col] == y_neg].shape[0]

    n_p_fav = cond_p_fav.shape[0]
    n_p_unfav = cond_p_unfav.shape[0]
    n_up_fav = cond_up_fav.shape[0]
    n_up_unfav = cond_up_unfav.shape[0]

    w_p_fav = n_fav * n_p / (num_samples * n_p_fav)
    w_p_unfav = n_unfav * n_p / (num_samples * n_p_unfav)
    w_up_fav = n_fav * n_up / (num_samples * n_up_fav)
    w_up_unfav = n_unfav * n_up / (num_samples * n_up_unfav)

    train_instance_weights = pd.DataFrame(np.ones(train.x.shape[0]), columns=["instance weights"])

    train_instance_weights.iloc[cond_p_fav.index] *= w_p_fav
    train_instance_weights.iloc[cond_p_unfav.index] *= w_p_unfav
    train_instance_weights.iloc[cond_up_fav.index] *= w_up_fav
    train_instance_weights.iloc[cond_up_unfav.index] *= w_up_unfav

    return train_instance_weights


def train_and_predict(train, test, classifier, C: float, kernel: str):
    """Train a logistic regression model and compute predictions on the given test data"""
    if classifier == "SVM":
        model = select_svm(C, kernel)
    else:
        model = LogisticRegression(solver="liblinear", random_state=888, max_iter=5000, C=C)
    model.fit(
        train.x, train.y.values.ravel(), sample_weight=compute_weights(train)["instance weights"]
    )
    return pd.DataFrame(model.predict(test.x), columns=["preds"])


def main():
    """This function runs the Kamiran&Calders method as a standalone program"""
    interface = InAlgoInterface()
    train, test = interface.load_data()
    classifier, C, kernel = interface.remaining_args()
    interface.save_predictions(train_and_predict(train, test, classifier, float(C), kernel))


if __name__ == "__main__":
    main()



"""
Implementation of Heaviside decision function
"""

import numpy
from .activation import Activation


class Heaviside(Activation):
    """Decision function that accepts predictions with score of 50% or above"""

    def apply(self, soft_output: numpy.ndarray) -> numpy.ndarray:
        def _heavi(x):
            return 1 if x >= 0.5 else 0

        return numpy.array([_heavi(x) for x in soft_output])

    def get_name(self) -> str:
        return "Heaviside"



"""
This module contains evaluators which apply algorithms over datasets and obtain metrics
"""


from . import evaluate_models, per_sensitive_attribute
from .parallelism import arrange_in_parallel, run_in_parallel



"""
Evaluator for a metric per sensitive attribute class
"""

from typing import Dict, List
import pandas as pd

from ethicml.utility.data_structures import DataTuple
from ..metrics.metric import Metric


class MetricNotApplicable(Exception):
    """Metric Not Applicable per sensitive attribute, apply to whole dataset instead"""


def metric_per_sensitive_attribute(
    predictions: pd.DataFrame, actual: DataTuple, metric: Metric
) -> Dict[str, float]:
    """Compute a metric repeatedly on subsets of the data that share a senstitive attribute"""
    if not metric.apply_per_sensitive:
        raise MetricNotApplicable(
            f"Metric {metric.name} is not applicable per sensitive "
            f"attribute, apply to whole dataset instead"
        )

    assert actual.s.shape[0] == actual.x.shape[0]
    assert actual.s.shape[0] == actual.y.shape[0]

    per_sensitive_attr: Dict[str, float] = {}

    s_columns: List[str] = [s_col for s_col in actual.s.columns]
    y_columns: List[str] = [y_col for y_col in actual.y.columns]
    pred_column: List[str] = [p_col for p_col in predictions.columns]
    assert len(y_columns) == 1

    for y_col in y_columns:
        for s_col in s_columns:
            for unique_s in actual.s[s_col].unique():
                for p_col in pred_column:
                    mask: pd.Series = (actual.s[s_col] == unique_s)
                    subset = DataTuple(
                        x=pd.DataFrame(
                            actual.x.loc[mask][actual.x.columns], columns=actual.x.columns
                        ).reset_index(drop=True),
                        s=pd.DataFrame(actual.s.loc[mask][s_col], columns=[s_col]).reset_index(
                            drop=True
                        ),
                        y=pd.DataFrame(actual.y.loc[mask][y_col], columns=[y_col]).reset_index(
                            drop=True
                        ),
                        name=actual.name,
                    )
                    pred_y = pd.DataFrame(
                        predictions.loc[mask][p_col], columns=[p_col]
                    ).reset_index(drop=True)
                    key = s_col + "_" + str(unique_s)
                    per_sensitive_attr[key] = metric.score(pred_y, subset)

    return per_sensitive_attr


def diff_per_sensitive_attribute(per_sens_res: Dict[str, float]) -> Dict[str, float]:
    """

    Args:
        per_sens_res:

    Returns:

    """
    sens_values = list(per_sens_res.keys())
    sens_values.sort()
    diff_per_sens = {}

    for i, _ in enumerate(sens_values):
        for j in range(i + 1, len(sens_values)):
            key: str = "{}-{}".format(sens_values[i], sens_values[j])
            i_value: float = per_sens_res[sens_values[i]]
            j_value: float = per_sens_res[sens_values[j]]
            diff_per_sens[key] = abs(i_value - j_value)

    return diff_per_sens


def ratio_per_sensitive_attribute(per_sens_res: Dict[str, float]) -> Dict[str, float]:
    """

    Args:
        per_sens_res:

    Returns:

    """
    sens_values = list(per_sens_res.keys())
    sens_values.sort()
    ratio_per_sens = {}

    for i, _ in enumerate(sens_values):
        for j in range(i + 1, len(sens_values)):
            key: str = "{}/{}".format(sens_values[i], sens_values[j])
            i_value: float = per_sens_res[sens_values[i]]
            j_value: float = per_sens_res[sens_values[j]]

            min_val = min(i_value, j_value)
            max_val = max(i_value, j_value)

            ratio_per_sens[key] = min_val / max_val

    return ratio_per_sens



"""
For assessing accuracy
"""

import pandas as pd
from sklearn.metrics import accuracy_score

from ethicml.utility.data_structures import DataTuple
from .metric import Metric


class Accuracy(Metric):
    """Accruacy"""

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        actual_y = actual.y.values
        return accuracy_score(actual_y, prediction)

    @property
    def name(self) -> str:
        return "Accuracy"



"""common variables / consts that make things run smoother"""

import os
from pathlib import Path

ROOT_DIR: str = os.path.abspath(os.path.join(os.path.abspath(__file__), os.pardir))
ROOT_PATH: Path = Path(__file__).parent.resolve()



"""Functions that are common to PyTorch models"""
import numpy as np
import torch
from torch.utils.data import Dataset

from ethicml.utility.data_structures import DataTuple, TestTuple


class TestDataset(Dataset):
    """Shared Dataset for pytorch models without labels"""

    def __init__(self, data: TestTuple):
        self.features = np.array(data.x.values, dtype=np.float32)
        self.sens_labels = np.array(data.s.values, dtype=np.float32)
        self.num = data.s.shape[0]
        self.s_size = data.s.shape[1]
        self.size = data.x.shape[1]
        self.x_names = data.x.columns
        self.s_names = data.s.columns

    def __getitem__(self, index):
        return self.features[index], self.sens_labels[index]

    def __len__(self):
        return self.num

    def names(self):
        return self.x_names, self.s_names


class CustomDataset(TestDataset):
    """Shared Dataset for pytorch models"""

    def __init__(self, data: DataTuple):
        super().__init__(data)
        self.class_labels = np.array(data.y.values, dtype=np.float32)
        self.y_size = data.y.shape[1]
        self.y_names = data.y.columns

    def __getitem__(self, index):
        return self.features[index], self.sens_labels[index], self.class_labels[index]

    def names(self):
        return self.x_names, self.s_names, self.y_names


def quadratic_time_mmd(data_first, data_second, sigma):
    """

    Args:
        data_first:
        data_second:
        sigma:

    Returns:

    """
    xx_gm = data_first @ data_first.t()
    xy_gm = data_first @ data_second.t()
    yy_gm = data_second @ data_second.t()
    x_sqnorms = torch.diagonal(xx_gm)
    y_sqnorms = torch.diagonal(yy_gm)

    pad_first = lambda x: torch.unsqueeze(x, 0)
    pad_second = lambda x: torch.unsqueeze(x, 1)

    gamma = 1 / (2 * sigma ** 2)
    # use the second binomial formula
    kernel_xx = torch.exp(-gamma * (-2 * xx_gm + pad_second(x_sqnorms) + pad_first(x_sqnorms)))
    kernel_xy = torch.exp(-gamma * (-2 * xy_gm + pad_second(x_sqnorms) + pad_first(y_sqnorms)))
    kernel_yy = torch.exp(-gamma * (-2 * yy_gm + pad_second(y_sqnorms) + pad_first(y_sqnorms)))

    xx_num = float(kernel_xx.shape[0])
    yy_num = float(kernel_yy.shape[0])

    mmd2 = (
        kernel_xx.sum() / (xx_num * xx_num)
        + kernel_yy.sum() / (yy_num * yy_num)
        - 2 * kernel_xy.sum() / (xx_num * yy_num)
    )
    return mmd2



"""
Wrapper for SKLearn implementation of MLP
"""
from typing import Optional, Tuple

from sklearn.neural_network import MLPClassifier

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmAsync
from ethicml.algorithms.inprocess.interface import conventional_interface
from ethicml.implementations import mlp


class MLP(InAlgorithmAsync):
    """MLP"""

    def __init__(
        self, hidden_layer_sizes: Optional[Tuple[int]] = None, activation: Optional[str] = None
    ):
        super().__init__()
        if hidden_layer_sizes is None:
            self.hidden_layer_sizes = MLPClassifier().hidden_layer_sizes
        else:
            self.hidden_layer_sizes = hidden_layer_sizes
        self.activation = MLPClassifier().activation if activation is None else activation

    def run(self, train, test):
        return mlp.train_and_predict(train, test, self.hidden_layer_sizes, self.activation)

    def _script_command(self, train_paths, test_paths, pred_path):
        script = ["-m", mlp.train_and_predict.__module__]
        args = conventional_interface(
            train_paths, test_paths, pred_path, str(self.hidden_layer_sizes), str(self.activation)
        )
        return script + args

    @property
    def name(self) -> str:
        return "MLP"



"""
This module contains metrics which can be applied to prediction results
"""
from .accuracy import Accuracy
from .metric import Metric
from .prob_pos import ProbPos
from .prob_neg import ProbNeg
from .confusion_matrix import confusion_matrix
from .normalized_mutual_information import NMI
from .tpr import TPR
from .tnr import TNR
from .ppv import PPV
from .npv import NPV
from .cv import CV
from .bcr import BCR
from .theil import Theil
from .prob_outcome import ProbOutcome



"""
Abstract Base Class for all datasets that come with the framework
"""

from abc import ABC, abstractmethod
from typing import Dict, List
from pathlib import Path

from ethicml.common import ROOT_PATH
from ethicml.data.util import filter_features_by_prefixes, get_discrete_features


class Dataset(ABC):
    """Base class for datasets"""

    _features: List[str]
    _class_label_prefix: List[str]
    _class_labels: List[str]
    _s_prefix: List[str]
    _sens_attrs: List[str]
    _cont_features: List[str]
    _disc_features: List[str]

    def __init__(self):
        self._features: List[str] = []
        self._class_label_prefix: List[str] = []
        self._class_labels: List[str] = []
        self._s_prefix: List[str] = []
        self._sens_attrs: List[str] = []
        self._cont_features: List[str] = []
        self._disc_features: List[str] = []
        self.features_to_remove: List[str] = []
        self.discrete_only: bool = False
        self._filepath: Path = ROOT_PATH / "data" / "csvs" / self.filename

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the dataset"""

    @property
    @abstractmethod
    def filename(self) -> str:
        """File from which to load the data"""

    @property
    def filepath(self) -> Path:
        return self._filepath

    @filepath.setter
    def filepath(self, loc: Path) -> None:
        self._filepath = loc

    @property
    def ordered_features(self) -> Dict[str, List[str]]:
        """
        Return a dictionary that has separate entries for the features, the labels and the
        sensitive attributes, but the x features are ordered so first are the discrete features,
        then the continuous
        """

        return {
            "x": self.discrete_features + self.continuous_features,
            "s": self._sens_attrs,
            "y": self._class_labels,
        }

    @property
    def feature_split(self) -> Dict[str, List[str]]:
        """
        Return a dictionary that has separate entries for the features, the labels and the
        sensitive attributes.
        """
        if self.discrete_only:
            self.features_to_remove += self.continuous_features

        return {
            "x": filter_features_by_prefixes(self.features, self.features_to_remove),
            "s": self._sens_attrs,
            "y": self._class_labels,
        }

    @property
    def continuous_features(self):
        """List of features that are continuous"""
        return self._cont_features

    @continuous_features.setter
    def continuous_features(self, feats: List[str]):
        self._cont_features = feats

    @property
    def features(self):
        """List of all features"""
        return self._features

    @features.setter
    def features(self, feats: List[str]):
        self._features = feats

    @property
    def s_prefix(self):
        """List of prefixes of the sensitive attribute"""
        return self._s_prefix

    @s_prefix.setter
    def s_prefix(self, sens_attrs: List[str]):
        self._s_prefix = sens_attrs
        self.features_to_remove += sens_attrs

    @property
    def sens_attrs(self):
        """List of sensitive attributes"""
        return self._sens_attrs

    @sens_attrs.setter
    def sens_attrs(self, sens_attrs: List[str]):
        self._sens_attrs = sens_attrs

    @property
    def class_labels(self):
        """List of class labels"""
        return self._class_labels

    @class_labels.setter
    def class_labels(self, labels: List[str]):
        self._class_labels = labels

    @property
    def class_label_prefix(self):
        """List of prefixes of class labels"""
        return self._class_label_prefix

    @class_label_prefix.setter
    def class_label_prefix(self, label_prefixs: List[str]):
        self._class_label_prefix = label_prefixs
        self.features_to_remove += label_prefixs

    @property
    def discrete_features(self) -> List[str]:
        """List of features that are discrete"""
        return get_discrete_features(
            self.features, self.features_to_remove, self.continuous_features
        )

    @discrete_features.setter
    def discrete_features(self, feats: List[str]):
        self._disc_features = feats
